"""
This package contains the various built-in plugins available for Chancy.
Plugins are used to extend the functionality of Chancy, while allowing for
significant customization.

Some Chancy plugins are enabled by default. You can pass your own into the
list of plugins to customize them or set ``no_default_plugins`` to ``True`` to
disable them completely:

.. code-block:: python

    async with Chancy(..., no_default_plugins=True, plugins=[]):
        ...

Default Plugins
---------------

- :class:`chancy.plugins.workflow.WorkflowPlugin`: This plugin provides a
  workflow system for managing complex job dependencies.
- :class:`chancy.plugins.metrics.Metrics`: This plugin collects metrics about
  the performance of the Chancy cluster and makes them available to the API,
  dashboard, and database.
- :class:`chancy.plugins.leadership.Leadership`: This plugin is responsible for
  managing leadership elections between multiple workers. Most other plugins
  require this plugin to be enabled.
- :class:`chancy.plugins.pruner.Pruner`: This plugin is responsible for pruning
  old jobs from the database when they are no longer needed.
- :class:`chancy.plugins.recovery.Recovery`: This plugin is responsible for
  recovering jobs that were interrupted during execution, such as when a
  worker crashes or is restarted.

Optional Plugins
----------------

Chancy also comes with some additional plugins that can be enabled as needed,
and typically require additional dependencies:

- :class:`chancy.plugins.api.Api`: This plugin provides an API and dashboard
  for viewing the state of the Chancy cluster.
- :class:`chancy.plugins.cron.Cron`: This plugin is responsible for running
  jobs to at recurring schedules, such as "once an hour" or "every Monday".
- :class:`chancy.plugins.retry.RetryPlugin`: This plugin provides a slightly
  more advanced job retry system with jitter, backoff, and other features.
- :class:`chancy.plugins.reprioritize.Reprioritize`: This plugin increases
  the priority of jobs based on how long they've been in the queue.
- :class:`chancy.plugins.sentry.SentryPlugin`: This plugin sends job exceptions
  to Sentry, along with metadata about the job and worker.

"""



from django.contrib.auth import aauthenticate
from django.contrib.auth.models import User
from starlette.authentication import AuthCredentials, BaseUser, SimpleUser
from starlette.requests import Request, HTTPConnection

from chancy.plugins.api import AuthBackend


class DjangoAuthBackend(AuthBackend):
    async def login(
        self, request: Request, username: str, password: str
    ) -> bool:
        user: User | None = await aauthenticate(
            username=username, password=password
        )
        if user is not None and user.is_superuser:
            request.session["username"] = username
            return True
        return False

    async def logout(self, request: Request) -> None:
        request.session.pop("username", None)

    async def authenticate(
        self, conn: HTTPConnection
    ) -> tuple[AuthCredentials, BaseUser] | None:
        username = conn.session.get("username")
        if username is not None:
            return AuthCredentials(["authenticated"]), SimpleUser(username)
        return None



import pytest

from chancy import job, Queue
from chancy.plugins.cron import Cron


@job()
def test_job():
    """Simple job function for testing"""
    pass


@pytest.mark.parametrize(
    "chancy",
    [{"plugins": [Cron(poll_interval=1)], "no_default_plugins": True}],
    indirect=True,
)
@pytest.mark.asyncio
async def test_schedule_job(chancy, worker):
    """Test scheduling a job with cron plugin"""
    await chancy.declare(Queue("default"))

    j = test_job.job.with_unique_key("test_job_cron")

    await Cron.schedule(chancy, "*/1 * * * *", j)

    schedules = await Cron.get_schedules(chancy, unique_keys=["test_job_cron"])

    assert len(schedules) == 1
    schedule = schedules["test_job_cron"]
    assert schedule["unique_key"] == "test_job_cron"
    assert schedule["cron"] == "*/1 * * * *"
    assert schedule["next_run"] is not None


@pytest.mark.parametrize(
    "chancy",
    [{"plugins": [Cron(poll_interval=1)], "no_default_plugins": True}],
    indirect=True,
)
@pytest.mark.asyncio
async def test_unschedule_job(chancy, worker):
    """Test unscheduling a job"""
    await chancy.declare(Queue("default"))

    j = test_job.job.with_unique_key("test_job_cron")
    await Cron.schedule(chancy, "*/5 * * * *", j)

    schedules = await Cron.get_schedules(chancy, unique_keys=["test_job_cron"])
    assert len(schedules) == 1

    await Cron.unschedule(chancy, "test_job_cron")

    schedules = await Cron.get_schedules(chancy, unique_keys=["test_job_cron"])
    assert len(schedules) == 0


@pytest.mark.parametrize(
    "chancy",
    [{"plugins": [Cron(poll_interval=1)], "no_default_plugins": True}],
    indirect=True,
)
@pytest.mark.asyncio
async def test_update_existing_job_schedule(chancy, worker):
    """Test updating an existing job's schedule"""
    await chancy.declare(Queue("default"))

    j = test_job.job.with_unique_key("test_job_cron")
    await Cron.schedule(chancy, "*/5 * * * *", j)
    await Cron.schedule(chancy, "*/10 * * * *", j)

    schedules = await Cron.get_schedules(chancy, unique_keys=["test_job_cron"])
    assert len(schedules) == 1
    schedule = schedules["test_job_cron"]
    assert schedule["cron"] == "*/10 * * * *"


@pytest.mark.parametrize(
    "chancy",
    [{"plugins": [Cron(poll_interval=1)], "no_default_plugins": True}],
    indirect=True,
)
@pytest.mark.asyncio
async def test_fail_scheduling_without_unique_key(chancy, worker):
    """Test that scheduling a job without a unique key fails"""
    await chancy.declare(Queue("default"))

    with pytest.raises(
        ValueError, match="requires that each job has a unique_key"
    ):
        await Cron.schedule(chancy, "*/5 * * * *", test_job)


@pytest.mark.parametrize(
    "chancy",
    [{"plugins": [Cron(poll_interval=1)], "no_default_plugins": True}],
    indirect=True,
)
@pytest.mark.asyncio
async def test_job_execution(chancy, worker):
    """Test that a scheduled job executes by verifying next_run is set"""
    await chancy.declare(Queue("default"))

    j = test_job.job.with_unique_key("immediate_job")

    await Cron.schedule(chancy, "*/1 * * * *", j)

    schedules = await Cron.get_schedules(chancy, unique_keys=["immediate_job"])
    assert len(schedules) == 1
    schedule = schedules["immediate_job"]
    assert schedule["next_run"] is not None


@pytest.mark.parametrize(
    "chancy",
    [{"plugins": [Cron(poll_interval=1)], "no_default_plugins": True}],
    indirect=True,
)
@pytest.mark.asyncio
async def test_get_schedules_all(chancy, worker):
    """Test getting all scheduled jobs"""
    await chancy.declare(Queue("default"))

    job1 = test_job.job.with_unique_key("test_job_1")
    job2 = test_job.job.with_unique_key("test_job_2")
    job3 = test_job.job.with_unique_key("test_job_3")

    await Cron.schedule(chancy, "*/5 * * * *", job1)
    await Cron.schedule(chancy, "*/10 * * * *", job2)
    await Cron.schedule(chancy, "*/15 * * * *", job3)

    all_schedules = await Cron.get_schedules(chancy)

    assert len(all_schedules) >= 3

    assert "test_job_1" in all_schedules
    assert "test_job_2" in all_schedules
    assert "test_job_3" in all_schedules


@pytest.mark.parametrize(
    "chancy",
    [{"plugins": [Cron(poll_interval=1)], "no_default_plugins": True}],
    indirect=True,
)
@pytest.mark.asyncio
async def test_get_schedules_filtered(chancy, worker):
    """Test getting scheduled jobs filtered by unique keys"""
    await chancy.declare(Queue("default"))

    job1 = test_job.job.with_unique_key("test_job_1")
    job2 = test_job.job.with_unique_key("test_job_2")
    job3 = test_job.job.with_unique_key("test_job_3")

    await Cron.schedule(chancy, "*/5 * * * *", job1)
    await Cron.schedule(chancy, "*/10 * * * *", job2)
    await Cron.schedule(chancy, "*/15 * * * *", job3)

    filtered_schedules = await Cron.get_schedules(
        chancy, unique_keys=["test_job_1", "test_job_3"]
    )

    assert len(filtered_schedules) == 2

    assert "test_job_1" in filtered_schedules
    assert "test_job_3" in filtered_schedules
    assert "test_job_2" not in filtered_schedules



"""
Unmanaged Django models that match the Chancy tables.

Makes the assumption that the Django default database is the same as the Chancy
database.
"""

__all__ = ("Job", "Worker", "Queue")

from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.conf import settings

from chancy.utils import chancy_uuid

PREFIX = getattr(settings, "CHANCY_PREFIX", "chancy_")


class Job(models.Model):
    id = models.UUIDField(primary_key=True, default=chancy_uuid, editable=False)
    queue = models.TextField()
    func = models.TextField()
    kwargs = models.JSONField(db_default="{}", blank=True)
    limits = models.JSONField(db_default="[]", blank=True)
    meta = models.JSONField(db_default="{}", blank=True)
    state = models.CharField(max_length=25, default="pending")
    priority = models.IntegerField(default=0)
    attempts = models.IntegerField(default=0)
    max_attempts = models.IntegerField(default=1)
    taken_by = models.TextField(null=True)
    created_at = models.DateTimeField(auto_now_add=True)
    started_at = models.DateTimeField(null=True)
    completed_at = models.DateTimeField(null=True)
    scheduled_at = models.DateTimeField(auto_now_add=True)
    unique_key = models.TextField(null=True)
    errors = models.JSONField(db_default="[]", blank=True)

    class Meta:
        managed = False
        db_table = f"{PREFIX}jobs"


class Worker(models.Model):
    worker_id = models.TextField(primary_key=True)
    tags = ArrayField(models.TextField(), default=list, blank=True)
    queues = ArrayField(models.TextField(), default=list, blank=True)

    last_seen = models.DateTimeField(auto_now=True)
    expires_at = models.DateTimeField()

    class Meta:
        managed = False
        db_table = f"{PREFIX}workers"


class Queue(models.Model):
    name = models.TextField(primary_key=True)
    state = models.TextField(default="active")
    concurrency = models.IntegerField(null=True)
    tags = ArrayField(models.TextField(), default=list)
    executor = models.TextField(
        default="chancy.executors.process.ProcessExecutor"
    )
    executor_options = models.JSONField(db_default="{}", blank=True)
    polling_interval = models.IntegerField(default=5)
    rate_limit = models.IntegerField(null=True, blank=True)
    rate_limit_window = models.IntegerField(null=True, blank=True)
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        managed = False
        db_table = f"{PREFIX}queues"



"""
Initial migration for the metrics plugin.
"""

from psycopg import AsyncCursor, sql
from psycopg.rows import DictRow

from chancy.migrate import Migration, Migrator


class MetricsInitialMigration(Migration):
    """
    Create the initial tables for the metrics plugin.
    """

    async def up(self, migrator: Migrator, cursor: AsyncCursor[DictRow]):
        """
        Create the metrics table.
        """
        await cursor.execute(
            sql.SQL(
                """
                CREATE TABLE IF NOT EXISTS {metrics_table} (
                    -- Primary key: metric_key + resolution + worker_id combination
                    metric_key VARCHAR(255) NOT NULL,
                    resolution VARCHAR(10) NOT NULL,  -- '1min', '5min', '1hour', etc.
                    worker_id VARCHAR(255) NOT NULL,
                    
                    -- Storage for time-series data
                    -- timestamp -> value mapping, stored in descending order (newest first)
                    timestamps TIMESTAMPTZ[] NOT NULL,
                    values JSONB[] NOT NULL,
                    
                    metric_type VARCHAR(20) NOT NULL,
                    updated_at TIMESTAMPTZ NOT NULL DEFAULT NOW(),
                    metadata JSONB NOT NULL DEFAULT '{{}}',
                    
                    PRIMARY KEY (metric_key, resolution, worker_id)
                )
                """
            ).format(metrics_table=sql.Identifier(f"{migrator.prefix}metrics"))
        )

        await cursor.execute(
            sql.SQL(
                """
                CREATE INDEX IF NOT EXISTS {metrics_updated_idx} 
                ON {metrics_table} (updated_at)
                """
            ).format(
                metrics_table=sql.Identifier(f"{migrator.prefix}metrics"),
                metrics_updated_idx=sql.Identifier(
                    f"{migrator.prefix}metrics_updated_idx"
                ),
            )
        )

    async def down(self, migrator: Migrator, cursor: AsyncCursor[DictRow]):
        """
        Drop the metrics table.
        """
        await cursor.execute(
            sql.SQL("DROP TABLE IF EXISTS {metrics_table}").format(
                metrics_table=sql.Identifier(f"{migrator.prefix}metrics")
            )
        )



import secrets

import click

from chancy import Chancy, Worker
from chancy.cli import run_async_command
from chancy.errors import MigrationsNeededError
from chancy.plugins.api.auth import SimpleAuthBackend


@click.group(name="worker")
def worker_group():
    """
    Worker management commands.
    """
    pass


@worker_group.command("start")
@click.option("--worker-id", "-w", help="The worker ID to use.")
@click.option(
    "--tags",
    "-t",
    help="Extra tags to apply to the worker.",
    multiple=True,
)
@click.pass_context
@run_async_command
async def worker_command(
    ctx: click.Context, worker_id: str | None, tags: list[str] | None
):
    """
    Start a worker.
    """
    chancy: Chancy = ctx.obj["app"]

    async with chancy:
        try:
            async with Worker(
                chancy,
                worker_id=worker_id,
                tags=set(tags) if tags else None,
            ) as worker:
                await worker.wait_for_shutdown()
        except MigrationsNeededError:
            click.echo(
                "The database is not up to date and is missing migrations.\n"
                "Please run `chancy misc migrate` to update the database.\n"
                "You can check the current migration status with"
                " `chancy misc check-migrations`."
            )
            return 1


@worker_group.command("web")
@click.option("--host", "-h", help="The host to bind to.", default="localhost")
@click.option("--port", "-p", help="The port to bind to.", default=8000)
@click.option(
    "--debug", "-d", help="Run the server in debug mode.", is_flag=True
)
@click.option(
    "--allow-origin",
    "-o",
    help="A list of allowed origins.",
    multiple=True,
    default=lambda: ["*"],
)
@click.pass_context
@run_async_command
async def web_command(
    ctx: click.Context,
    host: str,
    port: int,
    debug: bool,
    allow_origin: list[str],
):
    """
    Start the Chancy dashboard.
    """
    from chancy.plugins.api import Api

    chancy: Chancy = ctx.obj["app"]

    async with chancy:
        worker = Worker(chancy, tags=set())

        # The metrics plugin needs to be running to pull in cluster-wide
        # metrics.
        if metrics := chancy.plugins.get("chancy.metrics"):
            worker.manager.add("metrics", metrics.run(worker, chancy))

        if not (api := chancy.plugins.get("chancy.api")):
            chancy.log.info(
                "No API plugin was configured on the Chancy application,"
                " falling back to the default API."
            )
            auth = SimpleAuthBackend({"admin": secrets.token_urlsafe(32)})
            api = Api(
                host=host,
                port=port,
                allow_origins=allow_origin,
                debug=debug,
                authentication_backend=auth,
            )
            chancy.log.warning(
                f"No username or password was provided for the API, defaulting"
                f" to 'admin' with a random password: {auth.users['admin']}"
            )

        worker.manager.add("api", api.run(worker, chancy))
        await worker.wait_for_shutdown()



import asyncio
from typing import AsyncIterator

import pytest
import pytest_asyncio
import sys

from chancy import Chancy, Worker


@pytest.fixture(scope="session")
def event_loop_policy():
    # Since psycopg's asyncio implementation cannot use the default
    # proactor event loop on Windows, we need to use the selector event loop.
    if sys.platform == "win32":
        return asyncio.WindowsSelectorEventLoopPolicy()
    return asyncio.DefaultEventLoopPolicy()


@pytest_asyncio.fixture()
async def chancy(request):
    """
    Provides a Chancy application instance with an open connection pool
    to the test database.
    """
    async with Chancy(
        "postgresql://postgres:localtest@localhost:8190/postgres",
        **getattr(request, "param", {}),
    ) as chancy:
        await chancy.migrate()
        yield chancy
        await chancy.migrate(to_version=0)


@pytest.fixture
def chancy_just_app():
    """
    Provides just a configured chancy instance with no open connection pool
    or migrations.
    """
    return Chancy(
        "postgresql://postgres:localtest@localhost:8190/postgres",
    )


@pytest_asyncio.fixture()
async def worker(request, chancy) -> AsyncIterator[Worker]:
    """
    Starts and returns a Worker and the task associated with it.

    If the worker is not stopped by the time the test completes, it will be
    cancelled.
    """
    async with Worker(
        chancy, shutdown_timeout=60, **getattr(request, "param", {})
    ) as worker:
        yield worker


@pytest_asyncio.fixture()
async def worker_no_start(chancy) -> Worker:
    """
    Returns a Worker instance that has not been started.
    """
    return Worker(chancy)


@pytest.fixture(
    params=(
        [Chancy.Executor.Process, Chancy.Executor.Threaded]
        + (
            [Chancy.Executor.SubInterpreter]
            if sys.version_info >= (3, 13)
            else []
        )
    )
)
def sync_executor(request):
    """
    Provides a parameterized fixture for all sync executors.
    """
    return request.param


@pytest.fixture(params=[Chancy.Executor.Async])
def async_executor(request):
    """
    Provides a parameterized fixture for all async executors.
    """
    return request.param



import time
import asyncio

import pytest

from chancy import Chancy, Worker, Queue, QueuedJob, Reference, job, Job
from test_worker import job_that_fails


@job()
def slow_job_to_run():
    time.sleep(5)


@job()
def job_to_run():
    pass


@job()
async def async_job_to_run():
    pass


@job()
def job_with_instance(*, context: QueuedJob):
    context.meta["received_instance"] = True


@job()
async def async_job_with_instance(*, context: QueuedJob):
    context.meta["received_instance"] = True


@job()
async def very_long_job():
    await asyncio.sleep(60 * 60)


@job()
def sync_very_long_job():
    time.sleep(60)


@job(queue="low")
async def decorated_job_to_run():
    pass


@job(queue="low")
def sync_decorated_job_to_run():
    pass


@job()
def job_with_kwarg_generic(*, hello: list[str]):
    pass


@pytest.mark.asyncio
async def test_basic_job_sync(
    chancy: Chancy, worker: Worker, sync_executor: str
):
    """
    Ensures that a basic job can be run on all built-in executors.
    """
    await chancy.declare(Queue("low", executor=sync_executor))

    ref = await chancy.push(job_to_run.job.with_queue("low"))
    j = await chancy.wait_for_job(ref, timeout=30)

    assert j.state == j.State.SUCCEEDED


@pytest.mark.asyncio
async def test_basic_job_async(
    chancy: Chancy, worker: Worker, async_executor: str
):
    """
    Ensures that a basic job can be run on all built-in executors.
    """
    await chancy.declare(Queue("low", executor=async_executor))

    ref = await chancy.push(async_job_to_run.job.with_queue("low"))
    j = await chancy.wait_for_job(ref, timeout=30)

    assert j.state == j.State.SUCCEEDED


@pytest.mark.asyncio
async def test_async_job_on_sync_executor(
    chancy: Chancy, worker: Worker, sync_executor: str
):
    """
    Ensures that async jobs can run on sync executors.
    """
    await chancy.declare(Queue("low", executor=sync_executor))

    ref = await chancy.push(async_job_to_run.job.with_queue("low"))
    j = await chancy.wait_for_job(ref, timeout=30)

    assert j.state == j.State.SUCCEEDED


@pytest.mark.asyncio
async def test_wait_for_job_timeout(
    chancy: Chancy, worker: Worker, sync_executor: str
):
    """
    Ensures that waiting for a job times out as expected.
    """
    await chancy.declare(Queue("low", executor=sync_executor))

    ref = await chancy.push(job_to_run.job.with_queue("low"))
    j = await chancy.wait_for_job(ref, timeout=20)

    assert j.state == j.State.SUCCEEDED

    ref = await chancy.push(slow_job_to_run.job.with_queue("low"))
    with pytest.raises(asyncio.TimeoutError):
        await chancy.wait_for_job(ref, timeout=0.1)


@pytest.mark.asyncio
async def test_job_instance_kwarg(
    chancy: Chancy, worker: Worker, sync_executor: str
):
    """
    Test that jobs requesting a QueuedJob kwarg receive the correct instance.
    """
    await chancy.declare(Queue("low", executor=sync_executor))

    ref = await chancy.push(job_with_instance.job.with_queue("low"))
    j = await chancy.wait_for_job(ref, timeout=30)

    assert j.state == j.State.SUCCEEDED
    assert j.meta.get("received_instance") is True


@pytest.mark.asyncio
async def test_async_job_instance_kwarg(
    chancy: Chancy, worker: Worker, async_executor: str
):
    """
    Test that async jobs requesting a QueuedJob kwarg receive the correct
    instance.
    """
    await chancy.declare(Queue("low", executor=async_executor))

    ref = await chancy.push(async_job_with_instance.job.with_queue("low"))
    j = await chancy.wait_for_job(ref, timeout=30)

    assert j.state == j.State.SUCCEEDED
    assert j.meta.get("received_instance") is True


@pytest.mark.asyncio
async def test_async_job_instance_kwarg_on_sync_executor(
    chancy: Chancy, worker: Worker, sync_executor: str
):
    """
    Test that async jobs requesting a QueuedJob kwarg receive the correct
    instance when run on a sync executor.
    """
    await chancy.declare(Queue("low", executor=sync_executor))

    ref = await chancy.push(async_job_with_instance.job.with_queue("low"))
    j = await chancy.wait_for_job(ref, timeout=30)

    assert j.state == j.State.SUCCEEDED
    assert j.meta.get("received_instance") is True


@pytest.mark.asyncio
async def test_job_cancellation(chancy: Chancy, worker: Worker):
    """
    Test that jobs can be cancelled on supporting executors.
    """

    async def cancel_in_a_bit(to_cancel: Reference):
        await asyncio.sleep(10)
        await chancy.cancel_job(to_cancel)

    await chancy.declare(Queue("async", executor=Chancy.Executor.Async))
    await chancy.declare(Queue("sync", executor=Chancy.Executor.Process))

    ref = await chancy.push(very_long_job.job.with_queue("async"))
    asyncio.create_task(cancel_in_a_bit(ref))
    j = await chancy.wait_for_job(ref, timeout=30)
    assert j.state == j.State.FAILED

    ref = await chancy.push(sync_very_long_job.job.with_queue("sync"))
    asyncio.create_task(cancel_in_a_bit(ref))
    j = await chancy.wait_for_job(ref, timeout=30)
    assert j.state == j.State.FAILED


@pytest.mark.asyncio
async def test_job_signature_with_kwarg_marker(chancy, worker):
    """
    Ensures that a job with a kwarg-only marker and a generic can have its type
    signature checked.
    """
    await chancy.declare(Queue("low"))

    ref = await chancy.push(
        job_with_kwarg_generic.job.with_queue("low").with_kwargs(hello="world")
    )
    j = await chancy.wait_for_job(ref, timeout=30)
    assert j.state == j.State.SUCCEEDED


@pytest.mark.asyncio
async def test_failing_job(chancy: Chancy, worker: Worker):
    """
    Test that a job that fails will be marked as failed.
    """
    await chancy.declare(
        Queue(
            "default",
            concurrency=1,
        ),
        upsert=True,
    )
    ref = await chancy.push(Job.from_func(job_that_fails))
    j = await chancy.wait_for_job(ref, timeout=30)
    assert j.state == QueuedJob.State.FAILED



import asyncio

import pytest
from psycopg.rows import dict_row

from chancy import Chancy, Queue, Worker, job
from chancy.plugins.pruner import Pruner
from chancy.plugins.leadership import ImmediateLeadership


@job()
def job_to_run():
    pass


@pytest.mark.parametrize(
    "chancy",
    [
        {
            "plugins": [
                ImmediateLeadership(),
            ],
            "no_default_plugins": True,
        }
    ],
    indirect=True,
)
@pytest.mark.asyncio
async def test_pruner_functionality(chancy: Chancy, worker: Worker):
    """
    This test manually calls the prune method to avoid timing issues.
    """
    p = Pruner(Pruner.Rules.Queue() == "test_queue")
    await chancy.declare(Queue("test_queue"))

    ref = await chancy.push(job_to_run.job.with_queue("test_queue"))
    initial_job = await chancy.wait_for_job(ref)
    assert initial_job is not None, "Job should exist before pruning"

    async with chancy.pool.connection() as conn:
        async with conn.cursor(row_factory=dict_row) as cursor:
            await p.prune(chancy, cursor)

    pruned_job = await chancy.get_job(ref)
    assert pruned_job is None, "Job should be pruned"

    p = Pruner(
        (Pruner.Rules.Queue() == "test_queue") & (Pruner.Rules.Age() > 10)
    )
    ref = await chancy.push(job_to_run.job.with_queue("test_queue"))
    initial_job = await chancy.wait_for_job(ref)
    assert initial_job is not None, "Job should exist before pruning"

    async with chancy.pool.connection() as conn:
        async with conn.cursor(row_factory=dict_row) as cursor:
            await p.prune(chancy, cursor)

    not_pruned_job = await chancy.get_job(ref)
    assert not_pruned_job is not None, "Job should not be pruned yet"

    await asyncio.sleep(10)

    async with chancy.pool.connection() as conn:
        async with conn.cursor(row_factory=dict_row) as cursor:
            await p.prune(chancy, cursor)

    pruned_job = await chancy.get_job(ref)
    assert pruned_job is None, "Job should be pruned"



from psycopg import sql

from chancy.migrate import Migration


class ConvertToJSONB(Migration):
    """
    Convert JSON fields to JSONB for better performance and indexing capabilities.

    This migration updates all JSON fields in core tables to use JSONB type, which provides:
    - Better performance for reads
    - Ability to index JSON fields
    - More efficient storage
    """

    async def up(self, migrator, cursor):
        # Update the jobs table
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {jobs}
                ALTER COLUMN kwargs TYPE JSONB USING kwargs::JSONB,
                ALTER COLUMN limits TYPE JSONB USING limits::JSONB,
                ALTER COLUMN meta TYPE JSONB USING meta::JSONB,
                ALTER COLUMN errors TYPE JSONB USING errors::JSONB
                """
            ).format(jobs=sql.Identifier(f"{migrator.prefix}jobs"))
        )

        # Update the queues table
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {queues} 
                ALTER COLUMN executor_options TYPE JSONB USING 
                CASE WHEN executor_options IS NULL THEN NULL 
                ELSE executor_options::JSONB END
                """
            ).format(queues=sql.Identifier(f"{migrator.prefix}queues"))
        )

    async def down(self, migrator, cursor):
        # Revert the jobs table
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {jobs}
                ALTER COLUMN kwargs TYPE JSON USING kwargs::JSON,
                ALTER COLUMN limits TYPE JSON USING limits::JSON,
                ALTER COLUMN meta TYPE JSON USING meta::JSON,
                ALTER COLUMN errors TYPE JSON USING errors::JSON
                """
            ).format(jobs=sql.Identifier(f"{migrator.prefix}jobs"))
        )

        # Revert the queues table
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {queues} 
                ALTER COLUMN executor_options TYPE JSON USING 
                CASE WHEN executor_options IS NULL THEN NULL 
                ELSE executor_options::JSON END
                """
            ).format(queues=sql.Identifier(f"{migrator.prefix}queues"))
        )



import asyncio
import pytest

from chancy import Chancy, Queue, job
from chancy.plugins.workflow import Workflow, WorkflowPlugin, Sequence
from chancy.plugins.leadership import ImmediateLeadership
from chancy.utils import chancy_uuid


# Test jobs
@job()
def sync_success():
    return "success"


@job()
def sync_failure():
    raise ValueError("Failed job")


@job()
async def async_success():
    return "success"


@job()
async def async_failure():
    raise ValueError("Failed job")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "chancy",
    [
        {
            "plugins": [
                ImmediateLeadership(),
                WorkflowPlugin(),
            ],
            "no_default_plugins": True,
        },
    ],
    indirect=True,
)
async def test_sequential_workflow(chancy: Chancy, worker):
    """
    Test that a simple sequential workflow executes steps in order.
    """
    await chancy.declare(Queue("default"))

    workflow = (
        Workflow("sequential")
        .add("step1", sync_success)
        .add("step2", sync_success, ["step1"])
        .add("step3", sync_success, ["step2"])
    )

    workflow_id = await WorkflowPlugin.push(chancy, workflow)
    result = await WorkflowPlugin.wait_for_workflow(
        chancy, workflow_id, timeout=30
    )

    assert result.state == Workflow.State.COMPLETED
    assert len(result.steps) == 3
    for step in result.steps.values():
        assert step.state == step.state.SUCCEEDED


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "chancy",
    [
        {
            "plugins": [ImmediateLeadership(), WorkflowPlugin()],
            "no_default_plugins": True,
        },
    ],
    indirect=True,
)
async def test_parallel_workflow(chancy: Chancy, worker):
    """
    Test that parallel steps can execute concurrently.
    """
    await chancy.declare(Queue("default"))

    workflow = (
        Workflow("parallel")
        .add("setup", sync_success)
        .add_group(
            [
                ("parallel1", sync_success),
                ("parallel2", sync_success),
                ("parallel3", sync_success),
            ],
            ["setup"],
        )
        .add("finish", sync_success, ["parallel1", "parallel2", "parallel3"])
    )

    workflow_id = await WorkflowPlugin.push(chancy, workflow)
    result = await WorkflowPlugin.wait_for_workflow(
        chancy, workflow_id, timeout=30
    )

    assert result.state == Workflow.State.COMPLETED
    assert len(result.steps) == 5
    for step in result.steps.values():
        assert step.state == step.state.SUCCEEDED


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "chancy",
    [
        {
            "plugins": [ImmediateLeadership(), WorkflowPlugin()],
            "no_default_plugins": True,
        }
    ],
    indirect=True,
)
async def test_workflow_failure(chancy: Chancy, worker):
    """
    Test that workflow handles step failures correctly.
    """
    await chancy.declare(Queue("default"))

    workflow = (
        Workflow("failing")
        .add("step1", sync_success)
        .add("step2", sync_failure, ["step1"])
        .add("step3", sync_success, ["step2"])
    )

    workflow_id = await WorkflowPlugin.push(chancy, workflow)
    result = await WorkflowPlugin.wait_for_workflow(
        chancy, workflow_id, timeout=30
    )

    assert result.state == Workflow.State.FAILED
    assert result.steps["step1"].state == result.steps["step1"].state.SUCCEEDED
    assert result.steps["step2"].state == result.steps["step2"].state.FAILED
    # Step 3 was never queued due to the failure of a previous step.
    assert result.steps["step3"].state is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "chancy",
    [
        {
            "plugins": [ImmediateLeadership(), WorkflowPlugin()],
            "no_default_plugins": True,
        }
    ],
    indirect=True,
)
async def test_sequence(chancy: Chancy, worker):
    """
    Test that Sequence helper class works correctly.
    """
    await chancy.declare(Queue("default"))

    sequence = Sequence(
        "test_sequence",
        [
            sync_success,
            sync_success,
            sync_success,
        ],
    )

    workflow_id = await sequence.push(chancy)
    result = await WorkflowPlugin.wait_for_workflow(
        chancy, workflow_id, timeout=30
    )

    assert result.state == Workflow.State.COMPLETED
    assert len(result.steps) == 3
    for step in result.steps.values():
        assert step.state == step.state.SUCCEEDED


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "chancy",
    [
        {
            "plugins": [ImmediateLeadership(), WorkflowPlugin()],
            "no_default_plugins": True,
        }
    ],
    indirect=True,
)
async def test_workflow_timeout(chancy: Chancy, worker):
    """
    Test that wait_for_workflow respects timeout.
    """
    await chancy.declare(Queue("default"))

    @job()
    async def slow_job():
        await asyncio.sleep(5)

    workflow = Workflow("timeout").add("step1", slow_job)

    workflow_id = await WorkflowPlugin.push(chancy, workflow)

    with pytest.raises(asyncio.TimeoutError):
        await WorkflowPlugin.wait_for_workflow(chancy, workflow_id, timeout=0.1)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "chancy",
    [
        {
            "plugins": [ImmediateLeadership(), WorkflowPlugin()],
            "no_default_plugins": True,
        }
    ],
    indirect=True,
)
async def test_missing_workflow(chancy: Chancy, worker):
    """
    Test that waiting for a non-existent workflow raises KeyError.
    """
    with pytest.raises(KeyError):
        await WorkflowPlugin.wait_for_workflow(chancy, chancy_uuid(), timeout=1)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "chancy",
    [{"plugins": [ImmediateLeadership(), WorkflowPlugin()]}],
    indirect=True,
)
async def test_async_workflow(chancy: Chancy, worker):
    """
    Test that workflow handles async jobs correctly.
    """
    await chancy.declare(Queue("default", executor=Chancy.Executor.Async))

    workflow = (
        Workflow("async")
        .add("step1", async_success)
        .add("step2", async_success, ["step1"])
    )

    workflow_id = await WorkflowPlugin.push(chancy, workflow)
    result = await WorkflowPlugin.wait_for_workflow(
        chancy, workflow_id, timeout=30
    )

    assert result.state == Workflow.State.COMPLETED
    assert len(result.steps) == 2
    for step in result.steps.values():
        assert step.state == step.state.SUCCEEDED


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "chancy",
    [
        {
            "plugins": [ImmediateLeadership(), WorkflowPlugin()],
            "no_default_plugins": True,
        }
    ],
    indirect=True,
)
async def test_workflow_with_modified_jobs(chancy: Chancy, worker):
    """
    Test that workflow handles modified jobs correctly.
    """
    await chancy.declare(Queue("high_priority"))

    workflow = (
        Workflow("modified_jobs")
        .add(
            "step1",
            sync_success.job.with_queue("high_priority").with_priority(10),
        )
        .add(
            "step2",
            sync_success.job.with_queue("high_priority").with_max_attempts(3),
            ["step1"],
        )
    )

    workflow_id = await WorkflowPlugin.push(chancy, workflow)
    result = await WorkflowPlugin.wait_for_workflow(
        chancy, workflow_id, timeout=30
    )

    assert result.state == Workflow.State.COMPLETED
    assert len(result.steps) == 2
    for step in result.steps.values():
        assert step.state == step.state.SUCCEEDED



from psycopg import AsyncCursor, sql
from psycopg.rows import DictRow

from chancy.migrate import Migration, Migrator


class ConvertToJSONB(Migration):
    """
    Convert JSON fields to JSONB in cron table for better performance and indexing capabilities.

    This migration updates the job JSON field in cron table to use JSONB type, which provides:
    - Better performance for reads
    - Ability to index JSON fields
    - More efficient storage
    """

    async def up(self, migrator: Migrator, cursor: AsyncCursor[DictRow]):
        # Update cron table
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {cron}
                ALTER COLUMN job TYPE JSONB USING job::JSONB
                """
            ).format(cron=sql.Identifier(f"{migrator.prefix}cron"))
        )

    async def down(self, migrator: Migrator, cursor: AsyncCursor[DictRow]):
        # Revert cron table
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {cron}
                ALTER COLUMN job TYPE JSON USING job::JSON
                """
            ).format(cron=sql.Identifier(f"{migrator.prefix}cron"))
        )



from django.contrib import admin

from chancy.contrib.django.models import Job, Worker, Queue


class ReadOnlyAdmin(admin.ModelAdmin):
    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    # Oddly, this prevents the user from viewing the object.
    # def has_change_permission(self, request, obj=None):
    #     return False


@admin.register(Job)
class JobAdmin(admin.ModelAdmin):
    list_display = ("id", "queue", "func", "state", "attempts", "max_attempts")
    search_fields = ("id", "queue", "func")


@admin.register(Worker)
class WorkerAdmin(ReadOnlyAdmin):
    list_display = ("worker_id", "tags", "queues")
    search_fields = ("worker_id",)


@admin.register(Queue)
class QueueAdmin(admin.ModelAdmin):
    list_display = ("name", "state", "executor")
    search_fields = ("name",)
    list_filter = ("state",)



import dataclasses
from datetime import datetime, timedelta, timezone
import random
from typing import Any

from chancy.plugin import Plugin
from chancy.job import QueuedJob
from chancy.worker import Worker


class RetryPlugin(Plugin):
    """
    Plugin that handles job retries based on settings stored in the Job's
    metadata.

    This plugin can be used as an example for implementing your own retry
    policies.

    Usage:

    .. code-block:: python

        from chancy import job
        from chancy.plugins.retry import RetryPlugin

        @job()
        def job_that_fails():
            raise ValueError("This job should fail.")

        async with Chancy(..., plugins=[RetryPlugin()]) as chancy:
            await chancy.declare(Queue("default"))
            await chancy.push(
                job_that_fails.job.with_max_attempts(3).with_meta({
                    "retry_settings": {
                        "backoff": 2,
                        "backoff_factor": 3,
                        "backoff_limit": 300,
                        "backoff_jitter": [1, 5],
                    }
                })
            )

    The above example will retry the job 3 times, with a starting backoff of 2
    seconds, a backoff factor of 3, a backoff limit of 300 seconds, and a
    random jitter of between 1 and 5 seconds.
    """

    @staticmethod
    def get_identifier() -> str:
        return "chancy.retry_plugin"

    @staticmethod
    def calculate_next_run(job: QueuedJob, retry_settings: dict) -> datetime:
        delay = retry_settings.get("backoff", 1)
        factor = retry_settings.get("backoff_factor", 2.0)
        limit = retry_settings.get("backoff_limit", 300)
        jitter = retry_settings.get("backoff_jitter", [1, 5])

        delay *= factor ** (job.attempts - 1)
        delay = min(delay, limit)
        delay += random.uniform(*jitter)

        return datetime.now(timezone.utc) + timedelta(seconds=delay)

    async def on_job_completed(
        self,
        *,
        job: QueuedJob,
        worker: Worker,
        exc: Exception | None = None,
        result: Any = None,
    ) -> QueuedJob:
        if exc is None or job.state not in {
            job.State.FAILED,
            job.State.RETRYING,
        }:
            return job

        if job.attempts >= job.max_attempts:
            return job

        retry_settings = job.meta.get("retry_settings", {})

        # We don't need to adjust the # of attempts as the base executor will
        # do that for us.
        return dataclasses.replace(
            job,
            state=QueuedJob.State.RETRYING,
            scheduled_at=self.calculate_next_run(job, retry_settings),
            completed_at=None,
        )



import os
import asyncio
import threading
import functools
from concurrent.futures import Future

import sys
from typing import Any

from chancy.worker import Worker
from chancy.queue import Queue

try:
    # Only available in 3.14+
    from concurrent.futures import InterpreterPoolExecutor
except ImportError:
    # Available only in 3.13 with a backport.
    try:
        from interpreters_backport.concurrent.futures.interpreter import (
            InterpreterPoolExecutor,
        )
    except ImportError:
        raise ImportError(
            "The SubInterpreterExecutor requires Python 3.13 or later."
        )

from chancy.executors.base import Executor, ConcurrentExecutor
from chancy.job import QueuedJob, Limit


class SubInterpreterExecutor(ConcurrentExecutor):
    """
    .. note::

        This executor is experimental and may not work as expected. The
        sub-interpreter features it is built on is experimental and only
        available in Python 3.13 (with a backport) and >3.14.

        To use with 3.13, run:

        .. code-block:: bash

            pip install chancy[sub]

    To use this executor, simply pass the import path to this class in the
    ``executor`` field of your queue configuration or use the
    :class:`~chancy.app.Chancy.Executor` shortcut:

    .. code-block:: python

        async with Chancy("postgresql://localhost/postgres") as chancy:
            await chancy.declare(
                Queue(
                    name="default",
                    executor=Chancy.Executor.SubInterpreter,
                )
            )

    It's important to note that many C-based Python libraries are not (yet)
    compatible with sub-interpreters. If you encounter issues, you may need to
    switch to a different executor.

    :param worker: The worker instance associated with this executor.
    :param queue: The queue that this executor is associated with.
    """

    def __init__(self, worker: Worker, queue: Queue):
        super().__init__(worker, queue)
        self.pool = InterpreterPoolExecutor(
            max_workers=queue.concurrency,
            initializer=self.on_initialize_worker,
            initargs=(sys.path,),
        )

    @staticmethod
    def on_initialize_worker(parent_sys_path: list[str]):
        """
        This method is called in each worker before it begins running jobs.
        It can be used to perform any necessary setup, such as loading NLTK
        datasets or calling ``django.setup()``.

        By default, it replaces the running job's ``sys.path`` with the workers.
        """
        # Unlike all other executors, the InterpreterPoolExecutor does not
        # automatically inherit the parent process's sys.path. This is a
        # workaround to ensure that the worker has the same sys.path as the
        # parent process or tests will fail.
        sys.path = parent_sys_path

    async def push(self, job: QueuedJob) -> Future:
        job = await self.on_job_starting(job)
        future: Future = self.pool.submit(self.job_wrapper, job)
        self.jobs[future] = job
        future.add_done_callback(
            functools.partial(
                self._on_job_completed, loop=asyncio.get_running_loop()
            )
        )
        return future

    @classmethod
    def job_wrapper(cls, job: QueuedJob) -> tuple[QueuedJob, Any]:
        """
        This is the function that is actually started by the sub-interpreter
        executor. It's responsible for setting up necessary limits,
        running the job, and returning the result.
        """
        func, kwargs = Executor.get_function_and_kwargs(job)

        time_limit = next(
            (
                limit.value
                for limit in job.limits
                if limit.type_ == Limit.Type.TIME
            ),
            None,
        )

        timer = None
        if time_limit:
            timer = threading.Timer(time_limit, cls._timeout_handler)
            timer.start()

        try:
            if asyncio.iscoroutinefunction(func):
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                try:
                    result = loop.run_until_complete(func(**kwargs))
                finally:
                    loop.run_until_complete(loop.shutdown_asyncgens())
                    loop.close()
            else:
                result = func(**kwargs)
        finally:
            if timer:
                timer.cancel()

        return job, result

    @staticmethod
    def _timeout_handler():
        raise TimeoutError("Job timed out.")

    def _on_job_completed(
        self, future: Future, loop: asyncio.AbstractEventLoop
    ):
        job = self.jobs.pop(future)

        result = None
        exc = future.exception()
        if exc is None:
            job, result = future.result()

        asyncio.run_coroutine_threadsafe(
            self.on_job_completed(job=job, exc=exc, result=result),
            loop,
        )

    async def stop(self):
        self.pool.shutdown(cancel_futures=True)
        await super().stop()

    def get_default_concurrency(self) -> int:
        """
        Get the default concurrency level for this executor.

        This method is called when the queue's concurrency level is set to
        None. It should return the number of jobs that can be processed
        concurrently by this executor.

        On Python 3.13+, defaults to the number of logical CPUs on the system
        plus 4. On older versions of Python, defaults to the number of CPUs on
        the system plus 4. This mimics the behavior of Python's built-in
        ThreadPoolExecutor.
        """
        # Only available in 3.13+
        if hasattr(os, "process_cpu_count"):
            return min(32, (os.process_cpu_count() or 1) + 4)
        return min(32, (os.cpu_count() or 1) + 4)



from psycopg import sql

from chancy.migrate import Migration


class MakeMaxAttemptsNonNull(Migration):
    """
    Makes the max_attempts column non-null with a default value of 1, which
    matches the default value in the Job class.
    """

    async def up(self, migrator, cursor):
        # First set any existing NULL values to 1
        await cursor.execute(
            sql.SQL(
                """
                UPDATE {jobs}
                SET max_attempts = 1
                WHERE max_attempts IS NULL
                """
            ).format(jobs=sql.Identifier(f"{migrator.prefix}jobs"))
        )

        # Then alter the column to be non-null with a default
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {jobs}
                ALTER COLUMN max_attempts SET NOT NULL,
                ALTER COLUMN max_attempts SET DEFAULT 1
                """
            ).format(jobs=sql.Identifier(f"{migrator.prefix}jobs"))
        )

    async def down(self, migrator, cursor):
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {jobs}
                ALTER COLUMN max_attempts DROP NOT NULL,
                ALTER COLUMN max_attempts DROP DEFAULT
                """
            ).format(jobs=sql.Identifier(f"{migrator.prefix}jobs"))
        )



from django.db import models
from django.conf import settings

PREFIX = getattr(settings, "CHANCY_PREFIX", "chancy_")


class Cron(models.Model):
    unique_key = models.TextField(primary_key=True)
    job = models.JSONField(null=False)
    cron = models.TextField(null=False)
    last_run = models.DateTimeField()
    next_run = models.DateTimeField(null=False)

    class Meta:
        managed = False
        db_table = f"{PREFIX}cron"



"""
Metrics plugin for collecting and sharing metrics across workers.
"""

import asyncio
import datetime
import json
from collections import defaultdict
from dataclasses import dataclass, field
from itertools import groupby
from typing import Dict, List, Literal, Optional, Set, Tuple, Union, cast

from psycopg import sql
from psycopg.rows import dict_row

from chancy.app import Chancy
from chancy.job import QueuedJob
from chancy.plugin import Plugin
from chancy.worker import Worker

Resolution = Literal["1min", "5min", "1hour", "1day"]
MetricValue = Union[int, float, Dict[str, Union[int, float]]]
MetricPoint = Tuple[datetime.datetime, MetricValue]
MetricType = Literal["counter", "gauge", "histogram"]


@dataclass
class Metric:
    """
    A class representing a metric with values at different resolutions.
    """

    metric_type: MetricType
    values_1min: List[MetricPoint] = field(default_factory=list)
    values_5min: List[MetricPoint] = field(default_factory=list)
    values_1hour: List[MetricPoint] = field(default_factory=list)
    values_1day: List[MetricPoint] = field(default_factory=list)

    @property
    def values(self) -> Dict[Resolution, List[MetricPoint]]:
        """Get all values in a dictionary keyed by resolution."""
        return {
            "1min": self.values_1min,
            "5min": self.values_5min,
            "1hour": self.values_1hour,
            "1day": self.values_1day,
        }


class Metrics(Plugin):
    """
    A plugin that collects and aggregates various metrics from jobs and queues.

    The plugin maintains time-series data for various metrics, with automatic
    aggregation and pruning to keep storage requirements low while providing
    useful historical data.

    Metrics are synchronized across workers, so each worker has access to the
    full set of metrics.

    Example:

    .. code-block:: python

        from chancy import Chancy
        from chancy.plugins.metrics import Metrics

        async with Chancy(..., plugins=[Metrics()]) as chancy:
            ...

    The metrics are stored in a compact time-series format, with data points
    aggregated at different resolutions (1 minute, 5 minutes, 1 hour, 1 day).

    .. note::

        While you can use this plugin to record your own arbitrary metrics,
        it's not designed as a general-purpose monitoring solution. For more
        advanced monitoring and visualization, consider using a dedicated
        monitoring tool like Prometheus or Grafana.
    """

    def __init__(
        self,
        *,
        sync_interval: int = 60,
        max_points_per_resolution: Dict[Resolution, int] = None,
        maximum_metric_age: int = 60 * 60 * 24 * 90,
        collection_interval: int = 30,
    ):
        """
        Initialize the metrics plugin.

        :param sync_interval: How often to synchronize metrics with the
                              database and other workers.
        :param max_points_per_resolution: How many points to keep for each
                                          resolution.
        :param maximum_metric_age: The maximum age of metrics to keep in the
                                   database, in seconds. Defaults to 90 days.
        :param collection_interval: The interval at which to collect table size
                                    metrics, in seconds.
        """
        super().__init__()
        self.sync_interval = sync_interval

        self.worker_id = None

        # Default retention policy: how many points to keep for each resolution
        self.max_points: Dict[Resolution, int] = {
            "1min": 60,  # 1 hour of 1-minute data
            "5min": 288,  # 1 day of 5-minute data
            "1hour": 168,  # 1 week of hourly data
            "1day": 90,  # 90 days of daily data
        }

        if max_points_per_resolution:
            self.max_points.update(max_points_per_resolution)

        # In-memory cache of local metrics for this worker, updated in
        # real-time and synced to DB.
        self.local_metrics_cache: Dict[str, Metric] = {}

        # In-memory cache of aggregated metrics from all workers, updated on
        # pulls from DB.
        self.aggregated_metrics_cache: Dict[str, Metric] = {}

        # Track metrics that have been modified locally since last sync.
        self.modified_metrics: Set[str] = set()

        # Last sync timestamp.
        self.last_sync_time = datetime.datetime.now(datetime.timezone.utc)

        # Locks to prevent race conditions.
        self.metric_locks: Dict[str, asyncio.Lock] = {}

        # Maximum age of metrics to keep in the database
        self.maximum_metric_age = maximum_metric_age

        # The interval at which to collect table size metrics
        self.collection_interval = collection_interval

    async def run(self, worker: Worker, chancy: Chancy):
        """
        Run the metrics plugin.

        This continuously synchronizes metrics with the database and
        other workers.
        """
        # Store the worker_id for use in metrics storage/retrieval
        self.worker_id = worker.worker_id

        self.aggregated_metrics_cache.update(
            await self._get_raw_metrics(chancy)
        )

        # Start a task to collect table size metrics
        worker.manager.add(
            f"metrics_table_sizes_{self.worker_id}",
            self._collect_table_sizes(worker, chancy),
        )

        while await self.sleep(self.sync_interval):
            await self._sync_metrics(chancy)

    def migrate_package(self) -> str:
        """
        Get the package that contains the migrations for the metrics plugin.
        """
        return "chancy.plugins.metrics.migrations"

    def migrate_key(self) -> str:
        """
        Get the unique identifier for this plugin's migrations.
        """
        return "metrics"

    async def on_job_updated(self, *, worker: "Worker", job: QueuedJob):
        if job.started_at and job.completed_at:
            execution_time = (job.completed_at - job.started_at).total_seconds()
            await self.record_histogram_value(
                f"job:{job.func}:execution_time",
                execution_time,
            )
            await self.record_histogram_value(
                f"queue:{job.queue}:execution_time",
                execution_time,
            )

        state = job.state.value
        await self.increment_counter(f"job:{job.func}:{state}", 1)
        await self.increment_counter(f"global:status:{state}", 1)
        await self.increment_counter(f"queue:{job.queue}:throughput", 1)
        await self.increment_counter(f"queue:{job.queue}:{state}", 1)

    async def cleanup(self, chancy: Chancy) -> Optional[int]:
        """
        Clean up old metrics data.

        Called automatically by the Pruner plugin, or may be manually invoked.
        """

        async with chancy.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                query = sql.SQL(
                    """
                    DELETE
                    FROM {metrics_table}
                    WHERE
                        updated_at < NOW() - interval '{max_age} seconds'
                    """
                ).format(
                    metrics_table=sql.Identifier(f"{chancy.prefix}metrics"),
                    max_age=sql.Literal(self.maximum_metric_age),
                )

                await cursor.execute(query)
                return cursor.rowcount if cursor.rowcount > 0 else None

    async def _get_metric_lock(self, metric_key: str) -> asyncio.Lock:
        """Get a lock for a specific metric to prevent race conditions."""
        if metric_key not in self.metric_locks:
            self.metric_locks[metric_key] = asyncio.Lock()
        return self.metric_locks[metric_key]

    async def increment_counter(
        self, metric_key: str, value: Union[int, float]
    ) -> None:
        """
        Increment a counter metric.

        Counter metrics accumulate values over time periods.

        :param metric_key: The unique key for the metric
        :param value: The value to increment the counter by
        """
        now = datetime.datetime.now(datetime.timezone.utc)

        async with await self._get_metric_lock(metric_key):
            if metric_key not in self.local_metrics_cache:
                self.local_metrics_cache[metric_key] = Metric(
                    metric_type="counter"
                )

            metric = self.local_metrics_cache[metric_key]

            for resolution in self.max_points.keys():
                match resolution:
                    case "1min":
                        points = metric.values_1min
                    case "5min":
                        points = metric.values_5min
                    case "1hour":
                        points = metric.values_1hour
                    case "1day":
                        points = metric.values_1day

                bucket_time = self._get_bucket_time(now, resolution)

                if points and self._same_bucket(
                    points[0][0], bucket_time, resolution
                ):
                    current_value = cast(Union[int, float], points[0][1])
                    points[0] = (points[0][0], current_value + value)
                else:
                    points.insert(0, (bucket_time, value))

                if len(points) > self.max_points[resolution]:
                    points.pop()

                self.modified_metrics.add(metric_key)

    async def record_gauge(
        self, metric_key: str, value: Union[int, float]
    ) -> None:
        """
        Record a gauge metric which represents a point-in-time value.

        Gauge metrics record the most recent value in each time bucket.

        :param metric_key: The unique key for the metric
        :param value: The value to record
        """
        now = datetime.datetime.now(datetime.timezone.utc)

        async with await self._get_metric_lock(metric_key):
            if metric_key not in self.local_metrics_cache:
                self.local_metrics_cache[metric_key] = Metric(
                    metric_type="gauge"
                )

            metric = self.local_metrics_cache[metric_key]

            for resolution in self.max_points.keys():
                if resolution == "1min":
                    points = metric.values_1min
                elif resolution == "5min":
                    points = metric.values_5min
                elif resolution == "1hour":
                    points = metric.values_1hour
                elif resolution == "1day":
                    points = metric.values_1day

                bucket_time = self._get_bucket_time(now, resolution)

                if points and self._same_bucket(
                    points[0][0], bucket_time, resolution
                ):
                    points[0] = (points[0][0], value)
                else:
                    points.insert(0, (bucket_time, value))

                if len(points) > self.max_points[resolution]:
                    points.pop()

                self.modified_metrics.add(metric_key)

    async def record_histogram_value(
        self, metric_key: str, value: Union[int, float]
    ) -> None:
        """
        Record a value to a histogram metric.

        Histogram metrics track statistics (min, max, avg, count) for values
        over time periods.

        :param metric_key: The unique key for the metric
        :param value: The value to record
        """
        now = datetime.datetime.now(datetime.timezone.utc)

        async with await self._get_metric_lock(metric_key):
            if metric_key not in self.local_metrics_cache:
                self.local_metrics_cache[metric_key] = Metric(
                    metric_type="histogram"
                )

            metric = self.local_metrics_cache[metric_key]

            for resolution in self.max_points.keys():
                if resolution == "1min":
                    points = metric.values_1min
                elif resolution == "5min":
                    points = metric.values_5min
                elif resolution == "1hour":
                    points = metric.values_1hour
                elif resolution == "1day":
                    points = metric.values_1day

                bucket_time = self._get_bucket_time(now, resolution)

                if points and self._same_bucket(
                    points[0][0], bucket_time, resolution
                ):
                    current_stats = cast(
                        Dict[str, Union[int, float]], points[0][1]
                    )

                    count = cast(int, current_stats.get("count", 0)) + 1
                    current_sum = (
                        cast(float, current_stats.get("sum", 0)) + value
                    )
                    current_min = min(
                        cast(float, current_stats.get("min", value)), value
                    )
                    current_max = max(
                        cast(float, current_stats.get("max", value)), value
                    )

                    new_stats = {
                        "count": count,
                        "sum": current_sum,
                        "avg": current_sum / count,
                        "min": current_min,
                        "max": current_max,
                    }
                    points[0] = (points[0][0], new_stats)
                else:
                    initial_stats = {
                        "count": 1,
                        "sum": value,
                        "avg": value,
                        "min": value,
                        "max": value,
                    }
                    points.insert(0, (bucket_time, initial_stats))

                if len(points) > self.max_points[resolution]:
                    points.pop()

                self.modified_metrics.add(metric_key)

    async def _sync_metrics(self, chancy: Chancy) -> None:
        """
        Synchronize metrics with the database and other workers.
        """
        if self.modified_metrics:
            await self._push_metrics_to_db(chancy)

        self.aggregated_metrics_cache.update(
            await self._get_raw_metrics(chancy)
        )
        self.last_sync_time = datetime.datetime.now(datetime.timezone.utc)
        self.modified_metrics.clear()

    async def _push_metrics_to_db(self, chancy: Chancy) -> None:
        """
        Push modified metrics to the database.

        Only pushes metrics from the local_metrics_cache, not the aggregated
        cache.
        """
        if not self.modified_metrics:
            return

        metrics_to_insert = []

        for metric_key in self.modified_metrics:
            if metric_key not in self.local_metrics_cache:
                continue

            metric = self.local_metrics_cache[metric_key]

            for resolution in self.max_points.keys():
                if resolution == "1min":
                    points = metric.values_1min
                elif resolution == "5min":
                    points = metric.values_5min
                elif resolution == "1hour":
                    points = metric.values_1hour
                elif resolution == "1day":
                    points = metric.values_1day
                else:
                    continue

                if not points:
                    continue

                timestamps = [point[0] for point in points]
                values = [json.dumps(point[1]) for point in points]

                metrics_to_insert.append(
                    {
                        "metric_key": metric_key,
                        "resolution": resolution,
                        "worker_id": self.worker_id,
                        "timestamps": timestamps,
                        "values": values,
                        "metric_type": metric.metric_type,
                    }
                )

        if not metrics_to_insert:
            return

        async with chancy.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                await cursor.executemany(
                    sql.SQL(
                        """
                        INSERT INTO {metrics_table} (
                            metric_key,
                            resolution,
                            worker_id,
                            timestamps,
                            values,
                            metric_type,
                            updated_at
                        ) VALUES (
                            %(metric_key)s,
                            %(resolution)s,
                            %(worker_id)s, 
                            %(timestamps)s,
                            %(values)s,
                            %(metric_type)s,
                            NOW()
                        )
                        ON CONFLICT (
                            metric_key,
                            resolution,
                            worker_id
                        ) DO UPDATE
                        SET
                            timestamps = %(timestamps)s,
                            values = %(values)s,
                            metric_type = %(metric_type)s,
                            updated_at = NOW()
                    """
                    ).format(
                        metrics_table=sql.Identifier(f"{chancy.prefix}metrics")
                    ),
                    metrics_to_insert,
                )

    @staticmethod
    def _get_bucket_time(
        timestamp: datetime.datetime, resolution: Resolution
    ) -> datetime.datetime:
        """
        Get the appropriate time bucket for a given timestamp and resolution.
        """
        if resolution == "1min":
            return timestamp.replace(second=0, microsecond=0)
        elif resolution == "5min":
            minute = (timestamp.minute // 5) * 5
            return timestamp.replace(minute=minute, second=0, microsecond=0)
        elif resolution == "1hour":
            return timestamp.replace(minute=0, second=0, microsecond=0)
        elif resolution == "1day":
            return timestamp.replace(hour=0, minute=0, second=0, microsecond=0)
        else:
            raise ValueError(f"Unknown resolution: {resolution}")

    @staticmethod
    def _same_bucket(
        time1: datetime.datetime,
        time2: datetime.datetime,
        resolution: Resolution,
    ) -> bool:
        """
        Check if two timestamps belong to the same bucket for a given
        resolution.
        """
        if resolution == "1min":
            return (
                time1.year == time2.year
                and time1.month == time2.month
                and time1.day == time2.day
                and time1.hour == time2.hour
                and time1.minute == time2.minute
            )
        elif resolution == "5min":
            return (
                time1.year == time2.year
                and time1.month == time2.month
                and time1.day == time2.day
                and time1.hour == time2.hour
                and time1.minute // 5 == time2.minute // 5
            )
        elif resolution == "1hour":
            return (
                time1.year == time2.year
                and time1.month == time2.month
                and time1.day == time2.day
                and time1.hour == time2.hour
            )
        elif resolution == "1day":
            return (
                time1.year == time2.year
                and time1.month == time2.month
                and time1.day == time2.day
            )
        else:
            raise ValueError(f"Unknown resolution: {resolution}")

    def api_plugin(self) -> str | None:
        return "chancy.plugins.metrics.api.MetricsApiPlugin"

    def get_tables(self) -> list[str]:
        """Get the names of all tables this plugin is responsible for."""
        return ["metrics"]

    @staticmethod
    def get_identifier() -> str:
        return "chancy.metrics"

    async def _collect_table_sizes(self, worker: Worker, chancy: Chancy):
        """
        Collect size metrics for database tables.

        This runs as a separate task and collects table sizes every 30 seconds,
        but only when this worker is the leader to avoid duplicate metrics.
        """
        core_tables = [
            "jobs",
            "queues",
            "workers",
            "leader",
            "queue_rate_limits",
        ]

        while await self.sleep(self.collection_interval):
            await self.wait_for_leader(worker)

            plugin_tables = []
            for plugin in chancy.plugins.values():
                plugin_tables.extend(plugin.get_tables())

            all_tables = list(set(plugin_tables + core_tables))
            prefixed_tables = [
                f"{chancy.prefix}{table}" for table in all_tables
            ]

            async with chancy.pool.connection() as conn:
                async with conn.cursor(row_factory=dict_row) as cursor:
                    query = sql.SQL(
                        """
                        SELECT
                            table_name,
                            pg_total_relation_size(table_name) as total_size_bytes,
                            pg_relation_size(table_name) as table_size_bytes,
                            pg_indexes_size(table_name) as index_size_bytes
                        FROM unnest({tables}::text[]) AS table_name
                    """
                    ).format(tables=sql.Literal(prefixed_tables))

                    await cursor.execute(query)

                    async for result in cursor:
                        table_name = result["table_name"].removeprefix(
                            chancy.prefix
                        )

                        await self.record_histogram_value(
                            f"table:{table_name}:total_size_bytes",
                            result["total_size_bytes"],
                        )
                        await self.record_histogram_value(
                            f"table:{table_name}:table_size_bytes",
                            result["table_size_bytes"],
                        )
                        await self.record_histogram_value(
                            f"table:{table_name}:index_size_bytes",
                            result["index_size_bytes"],
                        )

    async def _get_raw_metrics(
        self,
        chancy: Chancy,
        *,
        metric_prefix: str | None = None,
        worker_id: str | None = None,
    ) -> Dict[str, Metric]:
        results = {}

        async with chancy.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                await cursor.execute(
                    sql.SQL(
                        """
                        SELECT 
                            metric_key,
                            resolution,
                            metric_type,
                            json_agg(timestamps ORDER BY worker_id, updated_at) as all_timestamps,
                            json_agg(values ORDER BY worker_id, updated_at) as all_values
                        FROM 
                            {metrics_table}
                        WHERE
                            (
                                %(worker_id)s::text IS NULL OR
                                    worker_id = %(worker_id)s
                            )
                        AND
                            (
                                %(metric_prefix)s::text IS NULL OR
                                    metric_key LIKE %(metric_prefix)s
                            )
                        GROUP BY 
                            metric_key, resolution, metric_type
                        ORDER BY 
                            metric_key, resolution
                        """
                    ).format(
                        metrics_table=sql.Identifier(f"{chancy.prefix}metrics"),
                    ),
                    {
                        "worker_id": worker_id,
                        "metric_prefix": (
                            f"{metric_prefix}%" if metric_prefix else None
                        ),
                    },
                )

                for metric_key, group in groupby(
                    [r async for r in cursor],
                    key=lambda r: r["metric_key"],
                ):
                    if metric_prefix and not self.matches_prefix(
                        metric_prefix, metric_key
                    ):
                        continue

                    group_list = list(group)

                    metric = results.setdefault(
                        metric_key,
                        Metric(metric_type=group_list[0]["metric_type"]),
                    )

                    for row in group_list:
                        resolution = cast(Resolution, row["resolution"])
                        merged_points = defaultdict(list)

                        for worker_idx, timestamps_array in enumerate(
                            row["all_timestamps"]
                        ):
                            values_array = row["all_values"][worker_idx]

                            for i in range(len(timestamps_array)):
                                timestamp = timestamps_array[i]
                                merged_points[timestamp].append(values_array[i])

                        result_points = []
                        for timestamp, values_list in merged_points.items():
                            if metric.metric_type == "counter":
                                merged_value = sum(values_list)
                            elif metric.metric_type == "gauge":
                                merged_value = values_list[0]
                            elif metric.metric_type == "histogram":
                                total_count = sum(
                                    v.get("count", 0) for v in values_list
                                )
                                total_sum = sum(
                                    v.get("sum", 0) for v in values_list
                                )

                                all_mins = [
                                    v.get("min")
                                    for v in values_list
                                    if "min" in v
                                ]
                                all_maxs = [
                                    v.get("max")
                                    for v in values_list
                                    if "max" in v
                                ]

                                merged_value = {
                                    "count": total_count,
                                    "sum": total_sum,
                                    "avg": (
                                        total_sum / total_count
                                        if total_count > 0
                                        else 0
                                    ),
                                    "min": min(all_mins) if all_mins else 0,
                                    "max": max(all_maxs) if all_maxs else 0,
                                }

                            result_points.append((timestamp, merged_value))

                        result_points.sort(key=lambda p: p[0], reverse=True)

                        max_points = self.max_points[resolution]
                        match resolution:
                            case "1min":
                                metric.values_1min = result_points[:max_points]
                            case "5min":
                                metric.values_5min = result_points[:max_points]
                            case "1hour":
                                metric.values_1hour = result_points[:max_points]
                            case "1day":
                                metric.values_1day = result_points[:max_points]

        return results

    async def get_metrics(
        self,
        chancy: Chancy,
        metric_prefix: str | None = None,
        worker_id: str | None = None,
    ) -> Dict[str, Metric]:
        """
        Get metrics matching the given prefix.

        If a worker_id is provided, only metrics for that worker are returned
        and are queried directly from the database. Otherwise, aggregated
        metrics are returned from the metrics cache.

        :param chancy: The Chancy application instance
        :param metric_prefix: Optional prefix to filter metrics by
        :param worker_id: Optional worker_id to filter metrics by
        """
        if worker_id and worker_id == self.worker_id:
            return {
                key: metric
                for key, metric in self.local_metrics_cache.items()
                if not metric_prefix or self.matches_prefix(metric_prefix, key)
            }
        elif worker_id:
            return await self._get_raw_metrics(
                chancy,
                metric_prefix=metric_prefix,
                worker_id=worker_id,
            )
        else:
            return {
                key: metric
                for key, metric in self.aggregated_metrics_cache.items()
                if not metric_prefix or self.matches_prefix(metric_prefix, key)
            }

    @staticmethod
    def matches_prefix(prefix: str, key: str) -> bool:
        """
        Check if the given key matches the given complete prefix.

        :param prefix: The prefix to match
        :param key: The key to check
        """
        key = key.split(":")
        prefix = prefix.split(":")
        return key[: len(prefix)] == prefix



from datetime import datetime, timezone, timedelta

from psycopg import sql
from psycopg.rows import dict_row

from chancy.plugin import Plugin, PluginScope
from chancy.app import Chancy
from chancy.worker import Worker


class Leadership(Plugin):
    """
    Leadership is used to ensure some plugins are run by only one worker at a
    time. For example, we wouldn't want every worker running database pruning
    every 60 seconds, as that would be immensely wasteful. Most other plugins
    require a Leadership plugin to be enabled.

    .. note::

        This plugin is enabled by default, you only need to provide it in the
        list of plugins to customize its arguments or if ``no_default_plugins``
        is set to ``True``.

    .. code-block:: python

        from chancy.plugins.leadership import Leadership

        async with Chancy(..., plugins=[
            Leadership()
        ]) as chancy:
            ...

    Implementation
    --------------
    The plugin implements a simple database-backed leader election system
    through a 'leader' table with the following characteristics:

    - Uses a single row with ID=1 to track the current leader worker.
    - Workers attempt to insert/update this row with their worker_id and an
      expiration timestamp.
    - Leadership is acquired when a worker successfully inserts the row (when
      no leader exists) or when the previous leader's entry has expired.
    - The leader renews its leadership by updating its expiration timestamp
      during each poll cycle.
    - If leadership renewal fails, the worker loses its leader status.
    - When a worker gains or loses leadership, appropriate notifications are
      sent via ``leadership.gained``, ``leadership.lost``, and
      ``leadership.renewed`` events.
    - Worker leadership status is tracked via a worker.is_leader event flag.

    :param poll_interval: The number of seconds between leadership poll
                          intervals.
    :param timeout: The number of seconds before a worker is considered to have
                    lost leadership.
    """

    def __init__(self, poll_interval: int = 60, timeout: int = 60 * 2):
        super().__init__()
        self.poll_interval = poll_interval
        self.timeout = timeout

        if self.timeout < self.poll_interval:
            raise ValueError(
                "The timeout must be greater than the poll interval. Typically,"
                " the timeout should be at least twice the poll interval."
            )

    @staticmethod
    def get_identifier() -> str:
        return "chancy.leadership"

    @classmethod
    def get_scope(cls) -> PluginScope:
        return PluginScope.WORKER

    def get_tables(self) -> list[str]:
        """Get the names of all tables this plugin is responsible for."""
        return ["leader"]

    async def run(self, worker: Worker, chancy: Chancy):
        prune_q = sql.SQL(
            """
            DELETE FROM {table} WHERE expires_at < %(now)s
            """
        ).format(table=sql.Identifier(f"{chancy.prefix}leader"))
        upsert_q = sql.SQL(
            """
            INSERT INTO {table} (id, worker_id, expires_at)
            VALUES (1, %(worker_id)s, %(expires_at)s)
            ON CONFLICT (id) DO UPDATE SET expires_at = %(expires_at)s
            """
        ).format(table=sql.Identifier(f"{chancy.prefix}leader"))
        insert_q = sql.SQL(
            """
            INSERT INTO {table} (id, worker_id, expires_at)
            VALUES (1, %(worker_id)s, %(expires_at)s) ON CONFLICT DO NOTHING
            """
        ).format(table=sql.Identifier(f"{chancy.prefix}leader"))

        # Trigger an immediate leadership check when a worker cleanly shuts
        # down.
        worker.hub.on(
            "worker.stopped",
            lambda worker_id: self.wakeup_signal.set(),
        )

        while await self.sleep(self.poll_interval):
            now = datetime.now(tz=timezone.utc)
            expires = now + timedelta(seconds=self.timeout)

            async with chancy.pool.connection() as conn:
                async with conn.cursor(row_factory=dict_row) as cur:
                    async with conn.transaction():
                        await cur.execute(prune_q, {"now": now})
                        if worker.is_leader.is_set():
                            await cur.execute(
                                upsert_q,
                                {
                                    "worker_id": worker.worker_id,
                                    "expires_at": expires,
                                },
                            )
                        else:
                            await cur.execute(
                                insert_q,
                                {
                                    "worker_id": worker.worker_id,
                                    "expires_at": expires,
                                },
                            )

                    is_leader = cur.rowcount == 1
                    was_leader = worker.is_leader.is_set()

                    if is_leader and was_leader:
                        chancy.log.debug(
                            f"Worker has renewed its leadership of the cluster"
                            f" until {expires}."
                        )
                        await chancy.notify(
                            cur,
                            "leadership.renewed",
                            {
                                "worker_id": worker.worker_id,
                            },
                        )
                    elif is_leader:
                        chancy.log.info(
                            f"Worker has become the leader of the cluster"
                            f" until {expires}."
                        )
                        worker.is_leader.set()
                        await chancy.notify(
                            cur,
                            "leadership.gained",
                            {
                                "worker_id": worker.worker_id,
                            },
                        )
                    elif was_leader:
                        chancy.log.info(
                            "Worker has lost leadership of the cluster."
                        )
                        worker.is_leader.clear()
                        await chancy.notify(
                            cur,
                            "leadership.lost",
                            {
                                "worker_id": worker.worker_id,
                            },
                        )


class ImmediateLeadership(Plugin):
    """
    A plugin that simply sets the worker as the leader immediately upon startup.

    This plugin is only ever intended for testing purposes, and should not be
    used in a production environment.
    """

    @classmethod
    def get_scope(cls) -> PluginScope:
        return PluginScope.WORKER

    @staticmethod
    def get_identifier() -> str:
        return "chancy.leadership"

    async def run(self, worker: Worker, chancy: Chancy):
        worker.is_leader.set()
        chancy.log.warning(
            "Worker has been set as leader without going through election."
            " This plugin should only be used for testing & debugging."
        )



"""
Unmanaged Django models for the workflow plugin.
"""

__all__ = ("Workflow", "WorkflowStep")

from django.db import models
from django.conf import settings

from chancy.utils import chancy_uuid

PREFIX = getattr(settings, "CHANCY_PREFIX", "chancy_")


class Workflow(models.Model):
    id = models.UUIDField(primary_key=True, default=chancy_uuid, editable=False)
    name = models.TextField(null=False)
    state = models.TextField(null=False)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        managed = False
        db_table = f"{PREFIX}workflows"


class WorkflowStep(models.Model):
    id = models.AutoField(primary_key=True)
    workflow = models.ForeignKey(
        Workflow,
        on_delete=models.CASCADE,
        related_name="steps",
        db_column="workflow_id",
    )
    step_id = models.TextField(null=False)
    job_data = models.JSONField(null=False)
    dependencies = models.JSONField(null=False)
    job_id = models.UUIDField(null=True, blank=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        managed = False
        db_table = f"{PREFIX}workflow_steps"
        unique_together = ("workflow", "step_id")



import pytest
import datetime

from chancy import Worker, Chancy, Queue, QueuedJob, job


low = Queue("low", concurrency=1)
high = Queue("high", concurrency=1)
default_concurrency = Queue("default_concurrency")


@job()
def job_to_run():
    return


@job()
def job_that_fails():
    raise ValueError("This job should fail.")


@job()
async def async_job_to_run():
    return


@pytest.mark.asyncio
async def test_multiple_queues(chancy: Chancy, worker: Worker):
    """
    Ensure that jobs can be pushed & retrieved from multiple enabled queues.
    """
    await chancy.declare(low)
    await chancy.declare(high)

    ref_low = await chancy.push(job_to_run.job.with_queue("low"))
    ref_high = await chancy.push(job_to_run.job.with_queue("high"))

    job_low = await chancy.wait_for_job(ref_low)
    job_high = await chancy.wait_for_job(ref_high)

    assert job_low.state == QueuedJob.State.SUCCEEDED
    assert job_high.state == QueuedJob.State.SUCCEEDED


@pytest.mark.asyncio
async def test_queue_listing(chancy: Chancy):
    """
    Ensure that we can retrieve a list of all the queues in the system.
    """
    await chancy.declare(low)
    await chancy.declare(high)

    queues = await chancy.get_all_queues()
    assert len(queues) == 2
    assert low in queues
    assert high in queues


@pytest.mark.asyncio
async def test_get_single_queue(chancy: Chancy):
    """
    Ensure that we can retrieve a single queue from the system.
    """
    await chancy.declare(low)
    await chancy.declare(high)

    queue = await chancy.get_queue("low")
    assert queue == low

    queue = await chancy.get_queue("high")
    assert queue == high


@pytest.mark.asyncio
async def test_default_concurrency_sync(
    chancy: Chancy, worker: Worker, sync_executor
):
    """
    Ensure that queues with default concurrency (None) can run jobs with sync executors.

    When concurrency is None, the executor's get_default_concurrency() method
    should determine the actual concurrency level.
    """
    queue = Queue("default_concurrency_sync", executor=sync_executor)
    await chancy.declare(queue)

    job_refs = []
    for _ in range(5):
        ref = await chancy.push(
            job_to_run.job.with_queue("default_concurrency_sync")
        )
        job_refs.append(ref)

    jobs = []
    for ref in job_refs:
        j = await chancy.wait_for_job(ref)
        jobs.append(j)

    for j in jobs:
        assert j.state == QueuedJob.State.SUCCEEDED


@pytest.mark.asyncio
async def test_default_concurrency_async(
    chancy: Chancy, worker: Worker, async_executor
):
    """
    Ensure that queues with default concurrency (None) can run jobs with async executors.

    When concurrency is None, the executor's get_default_concurrency() method
    should determine the actual concurrency level.
    """
    queue = Queue("default_concurrency_async", executor=async_executor)
    await chancy.declare(queue)

    job_refs = []
    for _ in range(5):
        ref = await chancy.push(
            async_job_to_run.job.with_queue("default_concurrency_async")
        )
        job_refs.append(ref)

    jobs = []
    for ref in job_refs:
        j = await chancy.wait_for_job(ref)
        jobs.append(j)

    for j in jobs:
        assert j.state == QueuedJob.State.SUCCEEDED


@pytest.mark.asyncio
async def test_queue_resume_time_tracking(chancy: Chancy):
    """
    Ensure that paused queues with resume_at set correctly track the time
    when they should automatically resume.
    """
    resume_time = datetime.datetime.min.replace(tzinfo=datetime.timezone.utc)
    paused_queue = Queue(
        "tracking_queue",
        concurrency=1,
        state=Queue.State.PAUSED,
        resume_at=resume_time,
    )

    await chancy.declare(paused_queue)
    queue = await chancy.get_queue("tracking_queue")
    assert queue.state == Queue.State.PAUSED
    assert queue.resume_at == resume_time

    await chancy.resume_queue("tracking_queue")
    queue = await chancy.get_queue("tracking_queue")
    assert queue.state == Queue.State.ACTIVE


@pytest.mark.asyncio
async def test_queue_auto_resume(chancy: Chancy, worker: Worker):
    """
    Ensure that paused queues with resume_at set in the past will automatically
    resume when processed by a worker.
    """
    paused_queue = Queue(
        "auto_resume_queue",
        concurrency=1,
        state=Queue.State.PAUSED,
        polling_interval=1,
    )

    await chancy.declare(paused_queue)

    # This job should never get processed since the queue is paused.
    ref = await chancy.push(job_to_run.job.with_queue("auto_resume_queue"))
    with pytest.raises(TimeoutError):
        await chancy.wait_for_job(ref, timeout=10)

    await chancy.pause_queue(
        "auto_resume_queue",
        resume_at=(
            datetime.datetime.now(tz=datetime.timezone.utc)
            + datetime.timedelta(seconds=10)
        ),
    )

    ref = await chancy.push(job_to_run.job.with_queue("auto_resume_queue"))
    j = await chancy.wait_for_job(ref, timeout=30)
    assert j.state == QueuedJob.State.SUCCEEDED

    # Same thing, but with a timedelta
    await chancy.pause_queue(
        "auto_resume_queue",
        resume_at=datetime.timedelta(seconds=10),
    )

    ref = await chancy.push(job_to_run.job.with_queue("auto_resume_queue"))
    j = await chancy.wait_for_job(ref, timeout=30)
    assert j.state == QueuedJob.State.SUCCEEDED



import os
import threading
from concurrent.futures import ThreadPoolExecutor, Future
import asyncio
import functools
from typing import Any

from chancy.executors.base import Executor, ConcurrentExecutor
from chancy.job import QueuedJob, Limit


class ThreadedExecutor(ConcurrentExecutor):
    """
    An Executor which uses a thread pool to run its jobs.

    This executor is useful for running I/O-bound jobs concurrently without the
    overhead of separate processes. It's not suitable for CPU-bound tasks due
    to Python's Global SubInterpreter Lock (GIL).

    When working with existing asyncio code, it's often easier and more
    efficient to use the :class:`~chancy.executors.asyncex.AsyncExecutor`
    instead, as it can run a very large number of jobs concurrently.

    To use this executor, simply pass the import path to this class in the
    ``executor`` field of your queue configuration or use the
    :class:`~chancy.app.Chancy.Executor` shortcut:

    .. code-block:: python

        async with Chancy("postgresql://localhost/postgres") as chancy:
            await chancy.declare(
                Queue(
                    name="default",
                    executor=Chancy.Executor.Threaded,
                )
            )

    :param worker: The worker instance associated with this executor.
    :param queue: The queue that this executor is associated with.
    """

    def __init__(self, worker, queue):
        super().__init__(worker, queue)
        self.pool = ThreadPoolExecutor(max_workers=queue.concurrency)

    async def push(self, job: QueuedJob) -> Future:
        job = await self.on_job_starting(job)
        future: Future = self.pool.submit(self.job_wrapper, job)
        self.jobs[future] = job
        future.add_done_callback(
            functools.partial(
                self._on_job_completed, loop=asyncio.get_running_loop()
            )
        )
        return future

    def job_wrapper(self, job: QueuedJob) -> tuple[QueuedJob, Any]:
        """
        This is the function that is actually started by the thread pool
        executor. It's responsible for setting up necessary limits,
        running the job, and returning the result.
        """
        func, kwargs = Executor.get_function_and_kwargs(job)

        time_limit = next(
            (
                limit.value
                for limit in job.limits
                if limit.type_ == Limit.Type.TIME
            ),
            None,
        )

        timer = None
        if time_limit:
            timer = threading.Timer(time_limit, self._timeout_handler)
            timer.start()

        try:
            if asyncio.iscoroutinefunction(func):
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                try:
                    result = loop.run_until_complete(func(**kwargs))
                finally:
                    loop.run_until_complete(loop.shutdown_asyncgens())
                    loop.close()
            else:
                result = func(**kwargs)
        finally:
            if timer:
                timer.cancel()

        return job, result

    @staticmethod
    def _timeout_handler():
        raise TimeoutError("Job timed out.")

    def _on_job_completed(
        self, future: Future, loop: asyncio.AbstractEventLoop
    ):
        job = self.jobs.pop(future)

        result = None
        exc = future.exception()
        if exc is None:
            job, result = future.result()

        asyncio.run_coroutine_threadsafe(
            self.on_job_completed(job=job, exc=exc, result=result),
            loop,
        )

    async def stop(self):
        self.pool.shutdown(cancel_futures=True)
        await super().stop()

    def get_default_concurrency(self) -> int:
        """
        Get the default concurrency level for this executor.

        This method is called when the queue's concurrency level is set to
        None. It should return the number of jobs that can be processed
        concurrently by this executor.

        On Python 3.13+, defaults to the number of logical CPUs on the system
        plus 4. On older versions of Python, defaults to the number of CPUs on
        the system plus 4. This mimics the behavior of Python's built-in
        ThreadPoolExecutor.
        """
        # Only available in 3.13+
        if hasattr(os, "process_cpu_count"):
            return min(32, (os.process_cpu_count() or 1) + 4)
        return min(32, (os.cpu_count() or 1) + 4)



from typing import Any

import sentry_sdk

from chancy.job import QueuedJob
from chancy.worker import Worker
from chancy.plugin import Plugin


class SentryPlugin(Plugin):
    """
    A plugin that sends errors to Sentry.

    This plugin will send all errors that occur during job execution to Sentry,
    along with extended metadata about the job such as the executing worker,
    the queue, and the job ID.

    This plugin assumes you've already configured sentry-sdk using
    ``sentry_sdk.init()`` at some point, to prevent conflicts with other Sentry
    integrations.

    Requires the sentry-sdk package version 2.0.0 or higher.

    Usage:

    .. code-block:: bash

        pip install sentry-sdk

    .. code-block:: python

        from chancy.plugins.sentry import SentryPlugin

        async with Chancy(..., plugins=[SentryPlugin()]) as chancy:
            pass
    """

    @staticmethod
    def get_identifier() -> str:
        return "chancy.sentry_plugin"

    async def on_job_completed(
        self,
        *,
        worker: Worker,
        job: QueuedJob,
        exc: Exception | None = None,
        result: Any | None = None,
    ) -> QueuedJob:
        if exc:
            with sentry_sdk.new_scope() as scope:
                scope.set_tags(
                    {
                        "chancy.func": job.func,
                        "chancy.worker": worker.worker_id,
                        "chancy.queue": job.queue,
                    }
                )
                scope.set_extra("chancy.job_id", job.id)
                scope.capture_exception(exc)
        return job



__all__ = ("SimpleAuthBackend", "AuthBackend")

import abc

from starlette.authentication import (
    AuthenticationBackend,
    AuthCredentials,
    BaseUser,
    SimpleUser,
)
from starlette.requests import HTTPConnection, Request


class AuthBackend(AuthenticationBackend, abc.ABC):
    @abc.abstractmethod
    async def login(
        self, request: Request, username: str, password: str
    ) -> bool:
        pass

    @abc.abstractmethod
    async def logout(self, request: Request) -> None:
        pass


class SimpleAuthBackend(AuthBackend):
    """
    A simple authentication backend that uses a dictionary of users and
    passwords.

    :param users: A dictionary of users and their passwords.
    """

    def __init__(self, users: dict[str, str]):
        self.users = users

    async def login(
        self, request: Request, username: str, password: str
    ) -> bool:
        if username in self.users and self.users[username] == password:
            request.session["username"] = username
            return True
        return False

    async def logout(self, request: Request) -> None:
        request.session.pop("username", None)

    async def authenticate(
        self, conn: HTTPConnection
    ) -> tuple[AuthCredentials, BaseUser] | None:
        username = conn.session.get("username")
        if username is None:
            return None

        return AuthCredentials(["authenticated"]), SimpleUser(username)



import pytest
from django.conf import settings


@pytest.fixture(scope="session")
def django_db_setup():
    settings.DATABASES["default"] = {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "postgres",
        "USER": "postgres",
        "PASSWORD": "localtest",
        "HOST": "localhost",
        "PORT": "8190",
    }



from psycopg import sql

from chancy.migrate import Migration


class AddResumeAtField(Migration):
    """
    Adds a resume_at datetime to the queues table.
    """

    async def up(self, migrator, cursor):
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {queues}
                ADD COLUMN resume_at TIMESTAMP WITH TIME ZONE
                """
            ).format(queues=sql.Identifier(f"{migrator.prefix}queues"))
        )

    async def down(self, migrator, cursor):
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {queues}
                DROP COLUMN resume_at
                """
            ).format(queues=sql.Identifier(f"{migrator.prefix}queues"))
        )



import asyncio
import enum
from dataclasses import dataclass, field
from datetime import datetime
from functools import partial
from typing import List, Dict, TextIO, Self
from psycopg import sql, AsyncCursor
from psycopg.rows import dict_row

from chancy.hub import Event
from chancy.plugin import Plugin
from chancy.app import Chancy
from chancy.worker import Worker
from chancy.job import Job, QueuedJob, IsAJob
from chancy.utils import json_dumps, chancy_uuid
from chancy.rule import Rule


@dataclass
class WorkflowStep:
    #: The job to execute when this step is ready.
    job: Job | IsAJob
    #: The unique ID of the step.
    step_id: str
    #: A list of step IDs that this step depends on.
    dependencies: List[str] = field(default_factory=list)
    #: The current state of the step.
    state: QueuedJob.State | None = QueuedJob.State.PENDING
    #: The unique ID of a running Job which is associated with this step.
    job_id: str | None = None


@dataclass
class Workflow:
    class State(enum.Enum):
        PENDING = "pending"
        RUNNING = "running"
        COMPLETED = "completed"
        FAILED = "failed"

    # A descriptive name for the workflow.
    name: str
    #: A dictionary of steps in the workflow, keyed by step ID.
    steps: Dict[str, WorkflowStep] = field(default_factory=dict)
    #: The current state of the workflow.
    state: State = State.PENDING

    #: The unique ID of a specific run of the workflow.
    id: str = field(default_factory=chancy_uuid)
    #: The time the workflow was created.
    created_at: datetime | None = None
    #: The time the workflow was last updated.
    updated_at: datetime | None = None

    def add(
        self, step_id: str, job: Job | IsAJob, dependencies: List[str] = None
    ) -> "Workflow":
        """
        Add a step to the workflow.

        .. code-block:: python

            workflow.add("step_1", job)
            workflow.add("step_2", job, ["step_1"])

        :param step_id: The ID of the step.
        :param job: The job to execute.
        :param dependencies: A list of step IDs that this step depends on.
        """
        self.steps[step_id] = WorkflowStep(
            job=job if isinstance(job, Job) else job.job,
            dependencies=dependencies or [],
            step_id=step_id,
        )
        return self

    def add_group(
        self,
        jobs: List[tuple[str, Job | IsAJob]],
        dependencies: List[str] = None,
    ) -> "Workflow":
        """
        Add a group of steps to the workflow.

        This is a convenience method for adding multiple steps to the workflow
        at once that are all dependent on the same set of dependencies.

        .. code-block:: python

            workflow = Workflow("my_workflow")
            workflow.add("setup", setup_job)
            workflow.add_group([
                ("step_1", job_1),
                ("step_2", job_2),
                ("step_3", job_3),
            ], ["setup"])

        :param jobs: A list of tuples of (step_id, job).
        :param dependencies: A list of step IDs that this step depends on.
        """
        for step_id, job in jobs:
            self.add(step_id, job, dependencies)
        return self

    def __repr__(self):
        return f"<Workflow({self.name!r}, {self.state!r})>"

    def __len__(self) -> int:
        return len(self.steps)

    def __iadd__(self, other: WorkflowStep) -> "Workflow":
        self.steps[other.step_id] = other
        return self

    def __iter__(self):
        return iter(self.steps.items())

    def __getitem__(self, key: str) -> WorkflowStep:
        return self.steps[key]

    def __delitem__(self, key: str):
        del self.steps[key]

    @property
    def is_complete(self) -> bool:
        return self.state in (self.State.COMPLETED, self.State.FAILED)

    @property
    def is_running(self) -> bool:
        return self.state == self.State.RUNNING

    @property
    def steps_by_state(self) -> Dict[QueuedJob.State, List[WorkflowStep]]:
        steps_by_state = {}
        for step in self.steps.values():
            steps_by_state.setdefault(step.state, []).append(step)
        return steps_by_state


class WorkflowPlugin(Plugin):
    """
    Support for dependency-based workflows.

    Workflows allow you to easily model complex processes that involve
    multiple jobs, each of which may depend on the completion of one or more
    other jobs. Workflows are modeled as a directed acyclic graph (DAG)
    where each node represents a job and each edge represents a dependency
    between two jobs.

    Workflows are implemented on top of the existing Chancy job system, meaning
    you can use all the existing job features, such as job retries, timeouts,
    scheduling, and so on.

    Enable the plugin by adding it to the list of plugins in the Chancy
    constructor:

    .. code-block:: python

        from chancy.plugins.leadership import Leadership
        from chancy.plugins.workflow import WorkflowPlugin

        async with Chancy(
            "postgresql://localhost/postgres",
            plugins=[Leadership(), WorkflowPlugin()]
        ) as chancy:
            ...

    Example
    -------

    We'll create a simple workflow runs the "top" job first, then the "left"
    and "right" jobs in parallel, and finally the "bottom" job:

    .. code-block:: python
       :caption: example_workflow.py

        import asyncio
        from chancy import Chancy, job
        from chancy.plugins.leadership import Leadership
        from chancy.plugins.workflow import Workflow, WorkflowPlugin

        @job()
        async def top():
            print(f"Top")

        @job()
        async def left():
            print(f"Left")

        @job()
        async def right():
            print(f"Right")

        @job()
        async def bottom():
            print(f"Bottom")

        async def main():
            async with Chancy(
                "postgresql://localhost/postgres",
                plugins=[Leadership(), WorkflowPlugin()]
            ) as chancy:
                workflow = (
                    Workflow("example")
                    .add("top", top)
                    .add("left", left, ["top"])
                    .add("right", right, ["top"])
                    .add("bottom", bottom, ["left", "right"])
                )
                await WorkflowPlugin.push(chancy, workflow)

        if __name__ == "__main__":
            asyncio.run(main())

    If we visualize our newly created workflow using :func:`generate_dot`, we
    get:

    .. graphviz:: _static/workflow.dot

    The full Workflow API is a little verbose if you just want to run a series
    of jobs in a specific order. In that case, you can use the
    :class:`Sequence` class to create a workflow from a list of jobs.

    Django Integration
    ------------------

    This plugin can be made available to the Django ORM and Admin interface.

    To enable this, you need to add the following to your Django settings:

    .. code-block:: python

        INSTALLED_APPS = [
            ...,
            "chancy.plugins.workflow.django",
        ]

    You can then query the workflows and steps using the Django ORM:

    .. code-block:: python

        from chancy.plugins.workflow.django.models import (
            Workflow,
            WorkflowStep
        )

        workflow = Workflow.objects.get(id="...")

        completed_steps = WorkflowStep.objects.filter(
            workflow=workflow,
            state=QueuedJob.State.SUCCEEDED
        )


    :param polling_interval: The interval at which to poll for new workflows.
    :param max_workflows_per_run: The maximum number of workflows to process
                                  in a single run of the plugin.
    """

    class Rules:
        class Age(Rule):
            def __init__(self):
                super().__init__("age")

            def to_sql(self) -> sql.Composable:
                return sql.SQL("EXTRACT(EPOCH FROM (NOW() - created_at))")

    def __init__(
        self,
        *,
        polling_interval: int = 30,
        max_workflows_per_run: int = 1000,
        pruning_rule: Rule = Rules.Age() > 60 * 60 * 24,
    ):
        super().__init__()
        self.polling_interval = polling_interval
        self.max_workflows_per_run = max_workflows_per_run
        self.pruning_rule = pruning_rule

    async def run(self, worker: Worker, chancy: Chancy):
        worker.hub.on(
            "workflow.created",
            lambda *args, **kwargs: self.wake_up(),
        )
        worker.hub.on(
            "workflow.step_completed",
            partial(
                self._on_single_step_completed, chancy=chancy, worker=worker
            ),
        )

        while await self.sleep(self.polling_interval):
            await self.wait_for_leader(worker)

            async with chancy.pool.connection() as conn:
                async with conn.cursor(row_factory=dict_row) as cursor:
                    # Grab the workflow IDs of workflows that are pending or
                    # running and lock them with FOR UPDATE.
                    await cursor.execute(
                        sql.SQL(
                            """
                            SELECT id
                            FROM {workflows} w
                            WHERE w.state IN ('pending', 'running')
                            FOR UPDATE SKIP LOCKED
                            LIMIT %(limit)s
                            """
                        ).format(
                            workflows=sql.Identifier(
                                f"{chancy.prefix}workflows"
                            ),
                        ),
                        {
                            "limit": self.max_workflows_per_run,
                        },
                    )

                    results = await cursor.fetchall()
                    if not results:
                        continue

                    workflows = await self.fetch_workflows_ex(
                        cursor,
                        chancy,
                        ids=[row["id"] for row in results],
                    )
                    for workflow in workflows:
                        if await self.process_workflow(
                            cursor, chancy, workflow
                        ):
                            await self.push_ex(cursor, chancy, workflow)

    async def _on_single_step_completed(
        self, event: Event, chancy: Chancy, worker: Worker
    ):
        """
        Called whenever a single step in a workflow is completed.

        Used to immediately process the workflow and check if it can be
        progressed to the next step without waiting for the next polling
        interval.
        """
        if not worker.is_leader:
            return

        async with chancy.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                await cursor.execute(
                    sql.SQL(
                        """
                        SELECT w.id
                        FROM {workflows} w
                        WHERE w.id = %(workflow_id)s
                        FOR UPDATE SKIP LOCKED
                        """
                    ).format(
                        workflows=sql.Identifier(f"{chancy.prefix}workflows"),
                        workflow_steps=sql.Identifier(
                            f"{chancy.prefix}workflow_steps"
                        ),
                    ),
                    {"workflow_id": event.body["workflow_id"]},
                )
                # Already locked by someone else.
                if not await cursor.fetchone():
                    return

                workflow = await self.fetch_workflow_ex(
                    cursor, chancy, event.body["workflow_id"]
                )
                if workflow is not None:
                    if await self.process_workflow(cursor, chancy, workflow):
                        await self.push_ex(cursor, chancy, workflow)

    async def on_job_updated(
        self,
        *,
        worker: "Worker",
        job: QueuedJob,
    ):
        if not job.meta.get("workflow_id"):
            return

        await worker.hub.emit(
            "workflow.step_completed", {"workflow_id": job.meta["workflow_id"]}
        )

    @classmethod
    async def fetch_workflow(cls, chancy: Chancy, id_: str) -> Workflow | None:
        """
        Fetch a single workflow from the database.

        :param chancy: The Chancy application.
        :param id_: The ID of the workflow to fetch.
        :return: The workflow, or None if it does not exist.
        """
        workflows = await cls.fetch_workflows(chancy, ids=[id_])
        return workflows[0] if workflows else None

    @classmethod
    async def fetch_workflow_ex(
        cls, cursor: AsyncCursor, chancy: Chancy, id_: str
    ) -> Workflow | None:
        """
        Fetch a single workflow from the database.

        This method is a lower-level version of fetch_workflow that accepts
        an existing cursor object, allowing it to be used in transactions.

        :param cursor: The cursor to use for the query.
        :param chancy: The Chancy application.
        :param id_: The ID of the workflow to fetch.
        :return: The workflow, or None if it does not exist.
        """
        workflows = await cls.fetch_workflows_ex(cursor, chancy, ids=[id_])
        return workflows[0] if workflows else None

    @classmethod
    async def fetch_workflows(
        cls,
        chancy: Chancy,
        *,
        states: list[str] | None = None,
        ids: list[str] | None = None,
        limit: int = 100,
    ) -> List[Workflow]:
        """
        Fetch workflows from the database, optionally matching the given
        conditions.

        :param chancy: The Chancy application.
        :param states: A list of states to match.
        :param ids: A list of IDs to match.
        :param limit: The maximum number of workflows to fetch.
        :return: A list of workflows.
        """
        async with chancy.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                return await cls.fetch_workflows_ex(
                    cursor,
                    chancy,
                    states=states,
                    ids=ids,
                    limit=limit,
                )

    @staticmethod
    async def fetch_workflows_ex(
        cursor: AsyncCursor,
        chancy: Chancy,
        *,
        states: list[str] | None = None,
        ids: list[str] | None = None,
        limit: int = 100,
    ) -> List[Workflow]:
        """
        Fetch workflows from the database, optionally matching the given
        conditions.

        This method is a lower-level version of fetch_workflows that accepts
        an existing cursor object, allowing it to be used in transactions.

        :param cursor: The cursor to use for the query.
        :param chancy: The Chancy application.
        :param states: A list of states to match.
        :param ids: A list of IDs to match.
        :param limit: The maximum number of workflows to fetch.
        :return: A list of workflows.
        """
        await cursor.execute(
            sql.SQL(
                """
                SELECT 
                    w.id, 
                    w.name, 
                    w.state,
                    w.created_at,
                    w.updated_at,
                    COALESCE(json_agg(
                        json_build_object(
                            'step_id', ws.step_id,
                            'job_data', ws.job_data,
                            'dependencies', ws.dependencies,
                            'state', j.state,
                            'job_id', ws.job_id
                        )
                    ), '[]'::json) as steps
                FROM {workflows} w
                LEFT JOIN {workflow_steps} ws ON w.id = ws.workflow_id
                LEFT JOIN {jobs} j ON ws.job_id = j.id
                WHERE (
                    %(states)s::text[] IS NULL OR
                    w.state = ANY(%(states)s::text[])
                ) AND (
                    %(ids)s::uuid[] IS NULL OR
                    w.id = ANY(%(ids)s::uuid[])
                )
                GROUP BY w.id, w.name, w.state
                LIMIT {limit}
                """
            ).format(
                workflows=sql.Identifier(f"{chancy.prefix}workflows"),
                workflow_steps=sql.Identifier(f"{chancy.prefix}workflow_steps"),
                jobs=sql.Identifier(f"{chancy.prefix}jobs"),
                limit=sql.Literal(limit),
            ),
            {
                "states": states,
                "ids": ids,
            },
        )
        rows = await cursor.fetchall()

        return [
            Workflow(
                id=row["id"],
                name=row["name"],
                state=Workflow.State(row["state"]),
                updated_at=row["updated_at"],
                created_at=row["created_at"],
                steps={
                    step["step_id"]: WorkflowStep(
                        job=Job.unpack(step["job_data"]),
                        dependencies=step["dependencies"],
                        state=(
                            QueuedJob.State(step["state"])
                            if step["state"]
                            else None
                        ),
                        step_id=step["step_id"],
                        job_id=step["job_id"],
                    )
                    for step in row["steps"]
                },
            )
            for row in rows
        ]

    @staticmethod
    async def process_workflow(
        cursor: AsyncCursor, chancy: Chancy, workflow: Workflow
    ) -> bool:
        """
        Process a single iteration of the given workflow, progressing the
        state of each step and the overall workflow as necessary.

        :param cursor: The cursor to use for the query.
        :param chancy: The Chancy application.
        :param workflow: The workflow to process.
        :return: True if the workflow was updated, False otherwise.
        """
        # If the workflow is already in a terminal state, there's no further
        # processing to do, although we may add future state handling here
        # for retries.
        if workflow.state in [Workflow.State.COMPLETED, Workflow.State.FAILED]:
            return False

        # Keep track of whether any changes were made to the workflow. If no
        # changes are made, we can skip updating the database.
        has_change = False
        starting_state = workflow.state

        # We check each step in the workflow to see:
        #    - If it has an associated job, and if so, what's the state of it?
        #    - If it has any dependencies, and if so, are they all completed?
        # If all dependencies are met, we can execute the job.
        for step_id, step in workflow.steps.items():
            # If the step is already in a terminal state, we can skip it.
            if step.state in [
                QueuedJob.State.SUCCEEDED,
                QueuedJob.State.FAILED,
            ]:
                continue

            dependencies = [workflow.steps[dep] for dep in step.dependencies]
            if all(
                dep.state == QueuedJob.State.SUCCEEDED for dep in dependencies
            ):
                if step.job_id is None:
                    step.job_id = (
                        await chancy.push_ex(
                            cursor,
                            step.job.with_meta(
                                {
                                    **step.job.meta,
                                    "workflow_id": str(workflow.id),
                                }
                            ),
                        )
                    ).identifier
                    has_change = True

        # Are all jobs complete, or any jobs failed? If so, we can mark the
        # workflow as completed or failed.
        states = workflow.steps_by_state
        if len(states.get(QueuedJob.State.SUCCEEDED, [])) == len(workflow):
            workflow.state = Workflow.State.COMPLETED
        elif states.get(QueuedJob.State.FAILED):
            workflow.state = Workflow.State.FAILED

        return starting_state != workflow.state or has_change

    @classmethod
    async def push(cls, chancy: Chancy, workflow: Workflow) -> str:
        """
        Push new workflow to the database.

        If the workflow already exists in the database, it will be updated
        instead.

        :param chancy: The Chancy application.
        :param workflow: The workflow to push.
        :return: The UUID of the newly created workflow.
        """
        async with chancy.pool.connection() as conn:
            async with conn.transaction():
                async with conn.cursor(row_factory=dict_row) as cursor:
                    return await cls.push_ex(cursor, chancy, workflow)

    @staticmethod
    async def push_ex(
        cursor: AsyncCursor, chancy: Chancy, workflow: Workflow
    ) -> str:
        """
        Push new workflow to the database.

        This method is a lower-level version of push that accepts an existing
        cursor object, allowing it to be used in transactions.

        If the workflow already exists in the database, it will be updated
        instead.

        :param cursor: The cursor to use for the query.
        :param chancy: The Chancy application.
        :param workflow: The workflow to push.
        :return: The UUID of the newly created workflow.
        """
        await cursor.execute(
            sql.SQL(
                """
                INSERT INTO {workflows} (
                    id,
                    name,
                    state,
                    created_at,
                    updated_at
                )
                VALUES (%s, %s, %s, NOW(), NOW())
                ON CONFLICT (id) DO UPDATE
                SET name = EXCLUDED.name,
                    state = EXCLUDED.state,
                    updated_at = NOW()
                RETURNING
                    id,
                    created_at,
                    updated_at,
                    (xmax = 0) as inserted
                """
            ).format(workflows=sql.Identifier(f"{chancy.prefix}workflows")),
            [workflow.id, workflow.name, workflow.state.value],
        )
        result = await cursor.fetchone()

        workflow.id = result["id"]
        workflow.created_at = result["created_at"]
        workflow.updated_at = result["updated_at"]
        inserted = result["inserted"]

        for step_id, step in workflow.steps.items():
            await cursor.execute(
                sql.SQL(
                    """
                    INSERT INTO {workflow_steps} (
                        workflow_id,
                        step_id,
                        job_data,
                        dependencies,
                        job_id
                    )
                    VALUES (%s, %s, %s, %s, %s)
                    ON CONFLICT (workflow_id, step_id) DO UPDATE
                    SET job_data = EXCLUDED.job_data,
                        dependencies = EXCLUDED.dependencies,
                        job_id = EXCLUDED.job_id,
                        updated_at = NOW()
                    """
                ).format(
                    workflow_steps=sql.Identifier(
                        f"{chancy.prefix}workflow_steps"
                    )
                ),
                [
                    workflow.id,
                    step_id,
                    json_dumps(step.job.pack()),
                    json_dumps(step.dependencies),
                    step.job_id,
                ],
            )

        await chancy.notify(
            cursor,
            f"workflow.{'created' if inserted else 'updated'}",
            {
                "id": workflow.id,
                "name": workflow.name,
            },
        )

        return workflow.id

    @staticmethod
    def generate_dot(workflow: Workflow, output: TextIO):
        """
        Generate a DOT file representation of the workflow.

        :param workflow: The Workflow object to visualize.
        :param output: A file-like object to write the DOT content to.
        """
        # Start the digraph
        output.write(f'digraph "{workflow.name}" {{\n')
        output.write("  rankdir=TB;\n")
        output.write(
            '  node [shape=box, style="rounded,filled", fontname="Arial"];\n'
        )

        # Define color scheme
        colors = {
            QueuedJob.State.PENDING: "lightblue",
            QueuedJob.State.RUNNING: "yellow",
            QueuedJob.State.SUCCEEDED: "lightgreen",
            QueuedJob.State.FAILED: "lightpink",
        }

        # Add nodes (steps)
        for step_id, step in workflow.steps.items():
            color = colors.get(step.state, "lightgray")
            output.write(
                f'  "{step_id}" [label="{step_id}\\n({step.state})",'
                f" fillcolor={color}];\n"
            )

        # Add edges (dependencies)
        for step_id, step in workflow.steps.items():
            for dep in step.dependencies:
                output.write(f'  "{dep}" -> "{step_id}";\n')

        # Add workflow info
        output.write('  labelloc="t";\n')
        output.write(
            f'  label="Workflow: {workflow.name}\\nState: {workflow.state}";\n'
        )

        # Close the digraph
        output.write("}\n")

    async def cleanup(self, chancy: Chancy) -> int | None:
        async with chancy.pool.connection() as conn:
            async with conn.cursor() as cursor:
                await cursor.execute(
                    sql.SQL(
                        """
                        DELETE FROM {workflows}
                        WHERE state NOT IN ('pending', 'running')
                        AND ({rule})
                        """
                    ).format(
                        workflows=sql.Identifier(f"{chancy.prefix}workflows"),
                        rule=self.pruning_rule.to_sql(),
                    )
                )
                return cursor.rowcount

    @classmethod
    async def wait_for_workflow(
        cls,
        chancy: Chancy,
        workflow_id: str,
        *,
        interval: int = 1,
        timeout: float | int | None = None,
    ) -> Workflow:
        """
        Wait for a workflow to complete.

        This method will loop until the workflow referenced by the provided ID
        has completed. The interval parameter controls how often the workflow
        status is checked. This will not block the event loop, so other tasks
        can run while waiting for the workflow to complete.

        Example
        -------

        .. code-block:: python

            workflow = Workflow("example")
            workflow.add("step1", job1)
            workflow.add("step2", job2, ["step1"])

            workflow_id = await WorkflowPlugin.push(chancy, workflow)
            completed_workflow = await WorkflowPlugin.wait_for_workflow(
                chancy,
                workflow_id,
                timeout=300  # 5 minute timeout
            )

        :param chancy: The Chancy application.
        :param workflow_id: The ID of the workflow to wait for.
        :param interval: The number of seconds to wait between checks.
        :param timeout: The maximum number of seconds to wait for the workflow to
            complete. If not provided, the method will wait indefinitely.
        :raises asyncio.TimeoutError: If the timeout is reached before the workflow
            completes.
        :raises KeyError: If the workflow does not exist.
        :return: The completed Workflow object.
        """
        async with asyncio.timeout(timeout):
            while True:
                workflow = await cls.fetch_workflow(chancy, workflow_id)
                if workflow is None:
                    raise KeyError(f"Workflow {workflow_id} not found")

                if workflow.is_complete:
                    return workflow

                await asyncio.sleep(interval)

    def migrate_package(self) -> str:
        return "chancy.plugins.workflow.migrations"

    def migrate_key(self) -> str:
        return "workflow"

    def api_plugin(self) -> str | None:
        return "chancy.plugins.workflow.api.WorkflowApiPlugin"

    def get_tables(self) -> list[str]:
        """Get the names of all tables this plugin is responsible for."""
        return ["workflows", "workflow_steps"]

    @staticmethod
    def get_identifier() -> str:
        return "chancy.workflow_plugin"

    @staticmethod
    def get_dependencies() -> list[str]:
        return ["chancy.leadership"]


class Sequence:
    """
    A sequential workflow.

    Sequences are a special case of workflows, where each step depends on the
    previous step. This forms a linear chain of jobs that are executed in
    order.

    Sequences are useful for defining sequences of jobs that must be executed
    in order, without the complexity of full workflows.

    Example
    -------

    .. code-block:: python
       :caption: example_sequence.py

        import asyncio
        from chancy import Chancy, job
        from chancy.plugins.workflow import Sequence

        @job()
        async def first():
            print("First")

        @job()
        async def second():
            print("Second")

        @job()
        async def third():
            print("Third")

        async def main():
            async with Chancy("postgresql://localhost/postgres") as chancy:
                sequence = Sequence("example_workflow", [first, second, third])
                await sequence.push(chancy)

        if __name__ == "__main__":
            asyncio.run(main())
    """

    def __init__(self, name: str, jobs: List[Job | IsAJob] = None):
        self.name = name
        self.jobs = jobs or []

    def add(self, job: Job | IsAJob) -> Self:
        """
        Add a job to the sequence.

        .. code-block:: python

            workflow = (
                Sequence("example_sequence")
                .add(first)
                .add(second)
                .add(third)
            )

        :param job: The job to add.
        """
        self.jobs.append(job)
        return self

    async def push(self, chancy: Chancy) -> str:
        """
        Push a sequence to the database.

        :param chancy: The Chancy application.
        :return: The UUID of the newly created chain.
        """
        workflow = Workflow(self.name)
        for i, job in enumerate(self.jobs):
            step_id = f"step_{i}"
            dependencies = [f"step_{i - 1}"] if i > 0 else []
            workflow.add(step_id, job, dependencies)

        return await WorkflowPlugin.push(chancy, workflow)



import asyncio
import code

import click
from psycopg import AsyncCursor
from psycopg.rows import DictRow, dict_row

from chancy import Chancy, Worker, Job, QueuedJob, Limit, Reference, Queue
from chancy.cli import run_async_command
from chancy.migrate import Migrator


@click.group(name="misc")
def misc_group():
    """
    Miscellaneous commands.
    """
    pass


@misc_group.command()
@click.option(
    "--to-version",
    "-t",
    help="The version to migrate to.",
    type=int,
)
@click.option(
    "--plugin",
    "-p",
    help="Migrate only the specified plugin.",
    type=str,
)
@click.pass_context
@run_async_command
async def migrate(
    ctx: click.Context, to_version: int | None, plugin: str | None
):
    """
    Migrate the database's core tables and any configured plugins up or down
    to the desired version.

    By default, migrates to the latest version, but can be controlled with
    the `--to-version` option.

    Use the `--plugin` option to migrate only a specific plugin instead of all plugins.
    """
    chancy: Chancy = ctx.obj["app"]

    async with chancy:
        if plugin:
            matching_plugins = [
                p
                for p in chancy.plugins.values()
                if p.__class__.__name__.lower() == plugin.lower()
            ]

            if not matching_plugins:
                raise click.ClickException(
                    f"No plugin found with name '{plugin}'"
                )

            if len(matching_plugins) > 1:
                raise click.ClickException(
                    f"Multiple plugins found with name '{plugin}'. Please use a more specific name."
                )

            await matching_plugins[0].migrate(chancy, to_version=to_version)
        else:
            await chancy.migrate(to_version=to_version)


async def _check_migrations(migrator: Migrator, cursor: AsyncCursor[DictRow]):
    """
    Check the migrations for a migrator.
    """
    good = "✓"
    bad = "✗"

    all_migrations = migrator.discover_all_migrations()
    current_version = await migrator.get_current_version(cursor)

    for i, migration in enumerate(all_migrations, 1):
        is_applied = i <= current_version
        click.echo(
            f"| |- [{good if is_applied else bad}] {migration.__class__.__name__} "
        )


@misc_group.command()
@click.pass_context
@run_async_command
async def check_migrations(ctx: click.Context):
    """
    Check the current migration status of the database.
    """
    chancy: Chancy = ctx.obj["app"]

    async with chancy:
        migrator = Migrator("chancy", "chancy.migrations", prefix=chancy.prefix)
        async with chancy.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                click.echo("Chancy Core")
                await _check_migrations(migrator, cursor)

                for plugin in chancy.plugins.values():
                    migrator = plugin.migrator(chancy)
                    if migrator is None:
                        continue

                    click.echo(f"|-{plugin.__class__.__name__} Plugin")
                    await _check_migrations(migrator, cursor)


@misc_group.command()
@click.pass_context
@run_async_command
async def shell(ctx):
    """
    Start an interactive shell with the Chancy app instance and common
    imports already available.
    """
    chancy: Chancy = ctx.obj["app"]

    async with chancy:
        code.interact(
            local={
                "chancy": chancy,
                "Worker": Worker,
                "Job": Job,
                "QueuedJob": QueuedJob,
                "Limit": Limit,
                "Reference": Reference,
                "Queue": Queue,
                "asyncio": asyncio,
            }
        )



__all__ = ("Api", "AuthBackend", "SimpleAuthBackend")
import os
import secrets
from pathlib import Path
from functools import partial
from typing import Type

import uvicorn
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.sessions import SessionMiddleware
from starlette.staticfiles import StaticFiles

from chancy import Worker, Chancy
from chancy.plugin import Plugin
from chancy.plugins.api.auth import AuthBackend, SimpleAuthBackend
from chancy.plugins.api.core import CoreApiPlugin
from chancy.plugins.api.plugin import ApiPlugin
from chancy.utils import import_string


class _SPAStaticFiles(StaticFiles):
    """
    A StaticFiles class that serves the SPA index.html file for any path that
    doesn't match an existing file.
    """

    def lookup_path(self, path: str) -> tuple[str, os.stat_result | None]:
        full_path, stat_result = super().lookup_path(path)
        if stat_result is None:
            return super().lookup_path("./index.html")
        return full_path, stat_result


class Api(Plugin):
    """
    Provides an API and a dashboard for viewing the state of the Chancy cluster.

    Running this plugin requires a few additional dependencies, you can install
    them with:

    .. code-block:: bash

        pip install chancy[web]

    To have the API run permanently on each Worker, you can add it to the
    plugins list when creating the Chancy instance:

    .. code-block:: python

        from chancy.plugins.api import Api
        from chancy.plugins.api.auth import SimpleAuthBackend

        async with Chancy(..., plugins=[
            Api(
                authentication_backend=SimpleAuthBackend({"admin": "password"}),
                secret_key="<a strong, random secret key>",
            ),
        ]) as chancy:
            ...

    .. note::

        The API is still mostly undocumented, as its development is driven by
        the needs of the dashboard and may change significantly before it
        becomes stable.

    Since it's very common to only want the dashboard temporarily, you can
    start it with the CLI. This can also be used to connect to a remote Chancy
    instance:

    .. code-block:: bash

        pip install chancy[cli,web]
        chancy --app worker.chancy worker web

    This will run the API and dashboard on port 8000 by default (you can change
    this with the ``--port`` and ``--host`` flags), and generate a temporary
    random password for the admin user.

    Screenshots
    -----------

    .. image:: ../misc/ux_jobs.png
        :alt: Jobs page

    .. image:: ../misc/ux_job_failed.png
        :alt: Failed job page

    .. image:: ../misc/ux_queue.png
        :alt: Queue page

    .. image:: ../misc/ux_workflow.png
        :alt: Worker page


    :param port: The port to listen on.
    :param host: The host to listen on.
    :param debug: Whether to run the server in debug mode.
    :param allow_origins: A list of origins that are allowed to access the API.
    """

    def __init__(
        self,
        *,
        port: int = 8000,
        host: str = "127.0.0.1",
        debug: bool = False,
        allow_origins: list[str] | None = None,
        secret_key: str | None = None,
        authentication_backend: AuthBackend,
    ):
        super().__init__()
        self.port = port
        self.host = host
        self.debug = debug
        self.root = Path(__file__).parent
        self.allow_origins = allow_origins or []
        self.plugins: set[Type[ApiPlugin]] = {CoreApiPlugin}
        self.authentication_backend = authentication_backend
        self.secret_key = secret_key or secrets.token_urlsafe(32)

    @staticmethod
    def get_identifier() -> str:
        return "chancy.api"

    async def run(self, worker: Worker, chancy: Chancy):
        """
        Start the web server.
        """
        for plug in chancy.plugins.values():
            api_plugin = plug.api_plugin()
            if api_plugin is None:
                continue

            api_plugin = import_string(api_plugin)
            if not issubclass(api_plugin, ApiPlugin):
                continue

            self.plugins.add(api_plugin)

        def _r(f):
            return partial(f, chancy=chancy, worker=worker)

        app = Starlette(
            debug=self.debug,
            middleware=[
                Middleware(
                    CORSMiddleware,
                    allow_origins=self.allow_origins,
                ),
                Middleware(SessionMiddleware, secret_key=self.secret_key),
                Middleware(
                    AuthenticationMiddleware,
                    backend=self.authentication_backend,
                ),
            ],
        )

        web_plugins = []
        # Look through all the enabled plugins for any that implement the
        # ApiPlugin interface. If they do, we merge them into our Starlette
        # app.
        for api_plugin in self.plugins:
            wp = api_plugin(self)
            web_plugins.append(wp)
            chancy.log.info(f"Loading API sub-plugin {wp.name()}")

            for route in wp.routes():
                if route.get("is_websocket"):
                    app.add_websocket_route(
                        route["path"],
                        _r(route["endpoint"]),
                        name=route["name"],
                    )
                else:
                    app.add_route(
                        route["path"],
                        _r(route["endpoint"]),
                        methods=route["methods"],
                        name=route["name"],
                    )

        # Add the wildcard route to the end of the list so that it doesn't
        # override any other routes and serves anything that should be handled
        # by the UI SPA.
        app.mount(
            "/",
            _SPAStaticFiles(
                packages=[("chancy.plugins.api", "dist")],
                html=True,
            ),
            name="ui",
        )

        server = uvicorn.Server(
            config=uvicorn.Config(
                app=app,
                host=self.host,
                port=self.port,
                log_level="error",
            )
        )

        chancy.log.info(f"Dashboard running at http://{self.host}:{self.port}")

        await server.serve()



from collections import defaultdict

from starlette.authentication import requires
from starlette.requests import Request
from starlette.responses import Response

from chancy.plugins.api.plugin import ApiPlugin
from chancy.utils import json_dumps


class MetricsApiPlugin(ApiPlugin):
    """
    API plugin for exposing metrics data.
    """

    def name(self):
        return "metrics"

    def routes(self):
        return [
            {
                "path": "/api/v1/metrics",
                "endpoint": self.get_metrics,
                "methods": ["GET"],
                "name": "get_metrics",
            },
            {
                "path": "/api/v1/metrics/{prefix}",
                "endpoint": self.get_metric_detail,
                "methods": ["GET"],
                "name": "get_metric_detail",
            },
        ]

    @staticmethod
    @requires(["authenticated"])
    async def get_metrics(request: Request, *, chancy, worker):
        """
        Get a list of all available metrics.
        """
        metrics_plugin = chancy.plugins["chancy.metrics"]

        metric_categories = defaultdict(list)

        metrics = await metrics_plugin.get_metrics(chancy)
        for key in metrics.keys():
            parts = key.split(":")
            category, metric_name = parts[0], parts[1]
            if metric_name not in metric_categories[category]:
                metric_categories[category].append(metric_name)

        return Response(
            json_dumps(
                {
                    "categories": metric_categories,
                    "count": len(metrics),
                }
            ),
            media_type="application/json",
        )

    @staticmethod
    @requires(["authenticated"])
    async def get_metric_detail(request: Request, *, chancy, worker):
        """
        Get detailed data for a specific metric.

        Can be filtered by worker_id with the worker_id query parameter.
        """
        metrics_plugin = chancy.plugins["chancy.metrics"]

        metric_prefix = request.path_params.get("prefix", "")
        resolution = request.query_params.get("resolution", "5min")
        worker_id = request.query_params.get("worker_id")

        metrics = await metrics_plugin.get_metrics(
            chancy,
            metric_prefix=metric_prefix,
            worker_id=worker_id,
        )

        return Response(
            json_dumps(
                {
                    metric_key: {
                        "data": [
                            {
                                "timestamp": timestamp,
                                "value": value,
                            }
                            for timestamp, value in getattr(
                                metric, f"values_{resolution}"
                            )
                        ],
                        "type": metric.metric_type,
                    }
                    for metric_key, metric in metrics.items()
                }
            ),
            media_type="application/json",
        )



"""
Tests for the Django models integration.
"""

import pytest

from chancy import job


@job()
def test_job():
    pass


@pytest.mark.django_db
@pytest.mark.asyncio
async def test_chancy_job_django_query(chancy, worker):
    """
    Test that a job created in Chancy can be queried using the Django ORM.
    """
    from chancy.contrib.django.models import Job

    j = await chancy.push(test_job)

    orm_job = await Job.objects.aget(id=j.identifier)

    assert orm_job.id == j.identifier
    assert orm_job.queue == "default"
    assert orm_job.func.endswith("test_job")


@pytest.mark.django_db
@pytest.mark.asyncio
async def test_chancy_worker_django_query(chancy, worker):
    """
    Test that we can query the Worker table.
    """
    from chancy.contrib.django.models import Worker

    orm_worker = await Worker.objects.aget(worker_id=worker.worker_id)

    assert orm_worker.worker_id == worker.worker_id
    assert len(orm_worker.tags) > 1 and "*" in orm_worker.tags


@pytest.mark.django_db
@pytest.mark.asyncio
async def test_chancy_queue_django_query(chancy, worker):
    """
    Test that we can query the Queue table.
    """
    from chancy.contrib.django.models import Queue

    orm_queue = await Queue.objects.aget(name="default")

    assert orm_queue.name == "default"
    assert orm_queue.state == "active"
    assert orm_queue.concurrency is None
    assert len(orm_queue.tags) == 1



from django.contrib import admin

from chancy.plugins.workflow.django.models import Workflow, WorkflowStep


class WorkflowStepInline(admin.TabularInline):
    model = WorkflowStep
    readonly_fields = (
        "id",
        "step_id",
        "dependencies",
        "job_id",
        "created_at",
        "updated_at",
    )
    extra = 0
    can_deletes = False


@admin.register(Workflow)
class WorkflowAdmin(admin.ModelAdmin):
    list_display = ("id", "name", "state", "created_at", "updated_at")
    search_fields = ("id", "name", "state")
    readonly_fields = ("id", "created_at", "updated_at")
    inlines = [WorkflowStepInline]


@admin.register(WorkflowStep)
class WorkflowStepAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "workflow",
        "step_id",
        "job_id",
        "created_at",
        "updated_at",
    )
    search_fields = (
        "id",
        "workflow__id",
        "workflow__name",
        "step_id",
        "job_id",
    )
    raw_id_fields = ("workflow",)
    readonly_fields = ("id", "created_at", "updated_at")



import pytest

from chancy import Chancy


@pytest.mark.django_db
@pytest.mark.asyncio
async def test_django_style_connection(settings):
    async with Chancy(settings.DATABASES["default"]) as app:
        await app.migrate()
        await app.migrate(to_version=0)



from starlette.authentication import requires
from starlette.responses import Response

from chancy.plugins.api import ApiPlugin
from chancy.plugins.cron import Cron
from chancy.utils import json_dumps


class CronApiPlugin(ApiPlugin):
    """
    Provides API endpoints for viewing cron jobs.
    """

    def name(self):
        return "cron"

    def routes(self):
        return [
            {
                "path": "/api/v1/crons",
                "endpoint": self.get_cron,
                "methods": ["GET"],
                "name": "get_cron",
            },
        ]

    @staticmethod
    @requires(["authenticated"])
    async def get_cron(request, *, chancy, worker):
        """
        Get all known cron jobs.
        """
        plugin: Cron = chancy.plugins["chancy.cron"]

        return Response(
            json_dumps(list((await plugin.get_schedules(chancy)).values())),
            media_type="application/json",
        )



from psycopg import AsyncCursor, sql
from psycopg.rows import DictRow

from chancy.migrate import Migration, Migrator


class ConvertToJSONB(Migration):
    """
    Convert JSON fields to JSONB in workflow tables for better performance and indexing capabilities.

    This migration updates all JSON fields in workflow tables to use JSONB type, which provides:
    - Better performance for reads
    - Ability to index JSON fields
    - More efficient storage
    """

    async def up(self, migrator: Migrator, cursor: AsyncCursor[DictRow]):
        # Update workflow_steps table
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {workflow_steps}
                ALTER COLUMN job_data TYPE JSONB USING job_data::JSONB,
                ALTER COLUMN dependencies TYPE JSONB USING dependencies::JSONB
                """
            ).format(
                workflow_steps=sql.Identifier(
                    f"{migrator.prefix}workflow_steps"
                )
            )
        )

    async def down(self, migrator: Migrator, cursor: AsyncCursor[DictRow]):
        # Revert workflow_steps table
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {workflow_steps}
                ALTER COLUMN job_data TYPE JSON USING job_data::JSON,
                ALTER COLUMN dependencies TYPE JSON USING dependencies::JSON
                """
            ).format(
                workflow_steps=sql.Identifier(
                    f"{migrator.prefix}workflow_steps"
                )
            )
        )



import time
import asyncio

import pytest

from chancy import Worker, Chancy, Job, Queue, QueuedJob
from chancy.errors import MigrationsNeededError


def job_to_run():
    return


def job_that_fails():
    raise ValueError("This job should fail.")


def job_that_sleeps():
    time.sleep(0.5)
    return


@pytest.mark.asyncio
async def test_queue_update(chancy: Chancy, worker: Worker):
    """
    Test that updating a queue's configuration successfully reconfigures the
    executor.
    """
    await chancy.declare(
        Queue(
            "test_update",
            concurrency=1,
            executor="chancy.executors.process.ProcessExecutor",
            polling_interval=5,
        ),
        upsert=True,
    )

    ref1 = await chancy.push(
        Job.from_func(job_that_sleeps, queue="test_update")
    )

    await asyncio.sleep(0.5)

    await chancy.declare(
        Queue(
            "test_update",
            concurrency=2,
            executor="chancy.executors.thread.ThreadedExecutor",
            polling_interval=2,
        ),
        upsert=True,
    )

    # The first job should still complete successfully
    job1 = await chancy.wait_for_job(ref1, timeout=10)
    assert job1.state == QueuedJob.State.SUCCEEDED

    # Push another job to the queue with the new configuration
    ref2 = await chancy.push(Job.from_func(job_to_run, queue="test_update"))
    job2 = await chancy.wait_for_job(ref2, timeout=10)
    assert job2.state == QueuedJob.State.SUCCEEDED

    # Check the executor has been updated in the worker
    assert (
        worker.executors["test_update"].__class__.__name__ == "ThreadedExecutor"
    )


@pytest.mark.asyncio
async def test_queue_removal(chancy: Chancy, worker: Worker):
    """
    Test that removing a queue properly cleans up its executor.
    """
    await chancy.declare(
        Queue(
            "test_removal",
            concurrency=1,
            polling_interval=1,
        ),
        upsert=True,
    )

    # Push a job and let it complete
    ref = await chancy.push(Job.from_func(job_to_run, queue="test_removal"))
    job = await chancy.wait_for_job(ref, timeout=30)
    assert job.state == QueuedJob.State.SUCCEEDED

    # Verify the executor exists
    assert "test_removal" in worker.executors

    await chancy.delete_queue("test_removal", purge_jobs=True)
    await worker.hub.wait_for("worker.queue.removed", timeout=30)

    # Give the executor time to clean up
    await asyncio.sleep(5)

    assert "test_removal" not in worker.executors


@pytest.mark.asyncio
async def test_error_on_needed_migrations(chancy_just_app):
    """
    Test that an error is raised if there are migrations that need to be
    applied before starting the worker.
    """
    with pytest.raises(MigrationsNeededError):
        async with chancy_just_app:
            async with Worker(chancy_just_app):
                pass



import abc
import enum
import asyncio
import typing
from typing import Any

from chancy import utils
from chancy.job import QueuedJob
from chancy.migrate import Migrator

if typing.TYPE_CHECKING:
    from chancy.app import Chancy
    from chancy.worker import Worker


class PluginScope(enum.Enum):
    #: The plugin is a general worker plugin.
    WORKER = "worker"


class Plugin(abc.ABC):
    """
    Base class for all plugins.

    Plugins are used to extend the functionality of the worker. When a worker
    starts, it will call :meth:`run` on all plugins that have a scope that
    matches the worker's scope.
    """

    def __init__(self):
        #: An asyncio.Event that can be used to wake up the plugin if it's
        #: sleeping.
        self.wakeup_signal = asyncio.Event()

    @classmethod
    def get_scope(cls) -> PluginScope:
        """
        Get the scope of this plugin. Scopes control when and where the plugin
        will be run.

        By default, plugins are scoped to the worker.
        """
        return PluginScope.WORKER

    async def run(self, worker: "Worker", chancy: "Chancy"):
        """
        Runs the plugin.

        This function can and should run indefinitely, as it will be cancelled
        when the worker is stopped.
        """

    def migrate_package(self) -> str | None:
        """
        Get the package name that contains the migration scripts for this
        plugin, if it has any.
        """

    def migrate_key(self) -> str | None:
        """
        Get the migration key for this plugin, if it has any.
        """

    async def migrate(self, chancy: "Chancy", *, to_version: int | None = None):
        """
        Migrate the database to the latest schema version.

        If `to_version` is provided, the database will be migrated to that
        specific version, up or down as necessary.
        """
        migrator = self.migrator(chancy)
        if migrator is None:
            return

        async with chancy.pool.connection() as conn:
            await migrator.migrate(conn, to_version=to_version)

    def migrator(self, chancy: "Chancy") -> Migrator | None:
        """
        Get a migrator for this plugin, if it has any migration scripts.
        """
        key = self.migrate_key()
        if key is None:
            return None

        package = self.migrate_package()
        if package is None:
            return None

        return Migrator(key, package, prefix=chancy.prefix)

    async def sleep(self, seconds: int) -> bool:
        """
        Sleep for a specified number of seconds, but allow the plugin to be
        woken up early.
        """
        await utils.sleep(seconds, events=[self.wakeup_signal.wait()])
        self.wakeup_signal.clear()
        return True

    @staticmethod
    async def wait_for_leader(worker: "Worker") -> None:
        """
        Wait until the worker running this plugin is the leader.
        """
        return await worker.is_leader.wait()

    def wake_up(self):
        """
        Wake up the plugin if it's sleeping.
        """
        self.wakeup_signal.set()

    async def cleanup(self, chancy: "Chancy") -> int | None:
        """
        Clean up any resources used by the plugin.

        Should return either None, if no work was done, or the number of
        rows cleaned up.

        .. note::

            Normally, you don't need to call this yourself. The Pruner plugin
            will call the cleanup method of all other registered plugins.
        """

    def api_plugin(self) -> str | None:
        """
        If this plugin has an associated API component, returns the import
        string for the plugin.
        """

    async def on_worker_started(self, worker: "Worker"):
        """
        Called when the worker has started.
        """

    async def on_job_starting(
        self,
        *,
        job: QueuedJob,
        worker: "Worker",
    ) -> QueuedJob:
        """
        Called when a job has been retrieved from the queue and is about to
        start.

        The passed job is immutable - to modify it, return a new QueuedJob
        with the desired changes.
        """
        raise NotImplementedError()

    async def on_job_completed(
        self,
        *,
        worker: "Worker",
        job: QueuedJob,
        exc: Exception | None = None,
        result: Any | None = None,
    ) -> QueuedJob:
        """
        Called after a job is completed (successfully or otherwise) and before
        the QueuedJob is updated in the database.

        If an exception occurred during the job, `exc` will be the exception
        instance instead of ``None``.

        The passed job is immutable - to modify it, return a new QueuedJob
        object with the desired changes.

        :param worker: The worker that is running the job.
        :param job: The job that was completed.
        :param exc: The exception that was raised, if any.
        :param result: The result of the job, if any.
        :return: The job to update in the database.
        """
        raise NotImplementedError()

    async def on_job_updated(
        self,
        *,
        worker: "Worker",
        job: QueuedJob,
    ):
        """
        Called after a job has been run and saved.

        Unlike on_job_completed, this method cannot modify the job, but the job
        is guaranteed to have been updated in the database by the time it is
        called.

        :param worker: The worker that is running the job.
        :param job: The job that was completed.
        """

    def get_tables(self) -> list[str]:
        """
        Get the names of all tables this plugin is responsible for.

        By default, returns an empty list.
        """
        return []

    @staticmethod
    @abc.abstractmethod
    def get_identifier() -> str:
        """
        Returns a unique identifier for this plugin.

        This identifier should be unique across all active plugins. If a
        custom plugin provides compatible functionality to a built-in plugin,
        it may use the same identifier as the built-in plugin.
        """

    @staticmethod
    def get_dependencies() -> list[str]:
        """
        Get the identifiers of all plugins this plugin depends on, if any.

        Plugins that depend on the presence of other plugins will refuse to
        start if those dependencies are not met.
        """
        return []

    def __repr__(self):
        return f"<{self.__class__.__name__}()>"



from django.contrib import admin

from chancy.plugins.cron.django.models import Cron


@admin.register(Cron)
class CronAdmin(admin.ModelAdmin):
    list_display = ("unique_key", "cron", "last_run", "next_run")
    search_fields = ("unique_key", "cron")
